import pickle
import tempfile
from pathlib import Path

import matplotlib.patches as mPatches
import matplotlib.pyplot as plt  # needed for visual inspection
import numpy as np
import pytest
from matplotlib.path import Path as mplPath
from shapely.geometry import MultiPolygon as shMultiPolygon
from shapely.geometry import Polygon as shPolygon

from locan import (
    AxisOrientedCuboid,
    AxisOrientedHypercuboid,
    Cuboid,
    Ellipse,
    EmptyRegion,
    Interval,
    MultiPolygon,
    Polygon,
    Rectangle,
    Region,
    Region1D,
    Region2D,
    Region3D,
    RegionND,
)
from locan.data.regions.region import _polygon_path


def test_Region():
    with pytest.raises(TypeError):
        Region()

    region = Region.from_intervals((0, 2))
    assert repr(region) == "Interval(0, 2)"
    region = Region.from_intervals(((0, 2), (0, 1)))
    assert repr(region) == "Rectangle((0, 0), 2, 1, 0)"
    region = Region.from_intervals(((0, 1), (0, 2), (0, 3)))
    assert repr(region) == "AxisOrientedCuboid((0, 0, 0), 1, 2, 3)"
    region = Region.from_intervals(((0, 1), (0, 2), (0, 3), (0, 4)))
    assert repr(region) == "AxisOrientedHypercuboid((0, 0, 0, 0), (1, 2, 3, 4))"


def test_Region1D():
    with pytest.raises(TypeError):
        Region1D()


def test_Region2D():
    with pytest.raises(TypeError):
        Region2D()
    points = ((2, 2), (2, 3), (3, 3), (3, 2.5), (2, 2))
    shapely_object = shPolygon(points)
    region = Region2D.from_shapely(shapely_object)
    assert isinstance(region, Polygon)

    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [
        ((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)),
        ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45)),
    ]
    shapely_object = shMultiPolygon([shapely_object, shPolygon(points, holes)])
    region = Region2D.from_shapely(shapely_object)
    assert isinstance(region, MultiPolygon)


def test_Region3D():
    with pytest.raises(TypeError):
        Region3D()


def test_RegionND():
    with pytest.raises(TypeError):
        RegionND()


def test_EmptyRegion():
    region = EmptyRegion()
    assert isinstance(region, Region)
    assert repr(region) == "EmptyRegion()"
    assert str(region) == "EmptyRegion()"
    assert region.dimension is None
    assert len(region.points) == 0
    assert region.centroid is None
    assert region.max_distance == 0
    assert region.region_measure == 0
    assert region.subregion_measure == 0
    assert region.bounds is None
    assert region.extent is None
    assert isinstance(region.bounding_box, EmptyRegion)
    other = Rectangle()
    assert isinstance(region.intersection(other), EmptyRegion)
    assert region.symmetric_difference(other) is other
    assert region.union(other) is other
    assert region.contains([[9, 8], [10.5, 10.5], [100, 100], [11, 12]]).size == 0
    assert region.contains([(10.5, 10)]).size == 0
    assert region.contains([(100, 100)]).size == 0
    assert region.contains([]).size == 0
    with pytest.raises(NotImplementedError):
        region.as_artist()
    assert isinstance(region.shapely_object, shPolygon)
    with pytest.raises(NotImplementedError):
        region.buffer(1)


def test_Interval():
    region = Interval()
    assert 2 not in region
    assert 0.5 in region
    assert isinstance(region, Region)
    assert repr(region) == "Interval(0, 1)"
    assert str(region) == "Interval(0, 1)"
    new_reg = eval(repr(region))
    assert isinstance(new_reg, Interval)
    with pytest.raises(AttributeError):
        region.lower_bound = None
        region.upper_bound = None
    assert region.dimension == 1
    assert np.array_equal(region.bounds, (0, 1))
    assert np.array_equal(region.intervals, (0, 1))
    assert region.extent == 1
    assert np.allclose(region.points.astype(float), (0, 1))
    assert region.centroid == 0.5
    assert region.max_distance == 1
    assert region.region_measure == 1
    assert region.subregion_measure == 0
    assert np.array_equal(region.contains((0, 0.5, 1, 2)), (0, 1))
    with pytest.raises(NotImplementedError):
        other = Interval()
        assert isinstance(region.intersection(other), EmptyRegion)
        assert region.symmetric_difference(other) is other
        assert region.union(other) is other
    assert region.contains((0.5,)) == (0,)
    assert region.contains((2,)).size == 0
    assert region.contains([]).size == 0
    # assert isinstance(region.as_artist(), mPatches)
    assert np.array_equal(region.buffer(1).points, [-1, 2])
    assert repr(region.buffer(1)) == "Interval(-1, 2)"

    region = Interval.from_intervals((0, 2))
    assert repr(region) == "Interval(0, 2)"
    region = Interval.from_intervals([0, 2])
    assert repr(region) == "Interval(0, 2)"
    region = Interval.from_intervals(np.array([0, 2]))
    assert repr(region) == "Interval(0, 2)"


def test_Rectangle():
    region = Rectangle((0, 0), 2, 1, 90)
    assert (10, 1) not in region
    assert (-0.5, 0.5) in region
    assert isinstance(region, Region)
    assert repr(region) == "Rectangle((0, 0), 2, 1, 90)"
    assert str(region) == "Rectangle((0, 0), 2, 1, 90)"
    new_reg = eval(repr(region))
    assert isinstance(new_reg, Rectangle)
    with pytest.raises(AttributeError):
        region.corner = None
        region.width = None
        region.height = None
        region.angle = None
    assert region.dimension == 2
    assert region.bounds == pytest.approx((-1, 0, 0, 2))
    assert np.array_equal(region.intervals, [(-1, pytest.approx(0)), (0, 2)])
    assert region.extent == pytest.approx((1, 2))
    assert len(region.points) == 5
    assert np.allclose(
        region.points.astype(float),
        [[0.0, 0.0], [-1.0, 0.0], [-1.0, 2.0], [0.0, 2.0], [0.0, 0.0]],
    )
    assert np.array_equal(region.centroid, (-0.5, 1))
    assert region.max_distance == np.sqrt(5)
    assert region.region_measure == 2
    assert region.subregion_measure == 6
    assert np.array_equal(
        region.contains([[0, 0], [-0.5, 0.5], [100, 100], [-1, 2]]), (1,)
    )
    other = Rectangle((0, 0), 2, 1, 0)
    assert isinstance(region.intersection(other), Polygon)
    assert isinstance(region.symmetric_difference(other), MultiPolygon)
    assert isinstance(region.union(other), Polygon)

    assert region.contains([(-0.5, 1)]) == (0,)
    assert region.contains([(10, 10)]).size == 0
    assert region.contains([]).size == 0
    assert isinstance(region.as_artist(), mPatches.Rectangle)
    assert isinstance(region.shapely_object, shPolygon)
    assert region.region_measure == region.shapely_object.area
    assert isinstance(region.buffer(1), Polygon)
    assert np.array_equal(region.bounding_box.corner, (-1.0, 0.0))
    assert region.bounding_box.width == pytest.approx(1)
    assert region.bounding_box.height == pytest.approx(2)

    region = Rectangle.from_intervals(((0, 2), (0, 1)))
    assert repr(region) == "Rectangle((0, 0), 2, 1, 0)"
    region = Rectangle.from_intervals([(0, 2), (0, 1)])
    assert repr(region) == "Rectangle((0, 0), 2, 1, 0)"
    region = Rectangle.from_intervals(np.array([(0, 2), (0, 1)]))
    assert repr(region) == "Rectangle((0, 0), 2, 1, 0)"


def test_Rectangle_visual():
    region = Rectangle((0, 0), 2, 1, 90)
    _fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.plot(*region.points.T, marker="o", color="Blue")
    ax.add_patch(region.as_artist(origin=(0, 0), fill=True, alpha=0.2))
    ax.plot(*np.array(region.shapely_object.exterior.coords).T, marker=".", color="Red")
    ax.plot(*region.centroid, "*", color="Green")
    ax.plot(*np.array(region.buffer(1).exterior.coords).T, marker=".", color="Yellow")
    region.plot(color="Green", alpha=0.2)
    # plt.show()
    plt.close("all")


def test_Ellipse():
    region = Ellipse((10, 10), 4, 2, 90)
    assert (1, 1) not in region
    assert (10, 10) in region
    assert isinstance(region, Region)
    assert repr(region) == "Ellipse((10, 10), 4, 2, 90)"
    assert str(region) == "Ellipse((10, 10), 4, 2, 90)"
    new_reg = eval(repr(region))
    assert isinstance(new_reg, Ellipse)
    with pytest.raises(AttributeError):
        region.center = None
        region.width = None
        region.height = None
        region.angle = None
    assert region.dimension == 2
    assert region.bounds == pytest.approx((9, 8, 11, 12))
    assert region.extent == pytest.approx((2, 4))
    assert len(region.points) == 65 or len(region.points) == 66
    assert np.array_equal(region.centroid, (10, 10))
    assert region.max_distance == 4
    assert region.region_measure == pytest.approx(6.283185307179586)
    assert region.subregion_measure == pytest.approx(9.688448216130086)
    assert np.array_equal(
        region.contains([[9, 8], [10.5, 10.5], [100, 100], [11, 12]]), (1,)
    )
    other = Rectangle((10, 10), 10, 10, 0)
    assert isinstance(region.intersection(other), Polygon)
    assert isinstance(region.symmetric_difference(other), MultiPolygon)
    assert isinstance(region.union(other), Polygon)
    assert region.contains([(10.5, 10)]) == (0,)
    assert region.contains([(100, 100)]).size == 0
    assert region.contains([]).size == 0
    assert isinstance(region.as_artist(), mPatches.Ellipse)
    assert isinstance(region.shapely_object, shPolygon)
    assert region.region_measure == pytest.approx(region.shapely_object.area, rel=10e-3)
    assert isinstance(region.buffer(1), Polygon)
    assert np.array_equal(region.bounding_box.corner, (9, 8))
    assert region.bounding_box.width == pytest.approx(2)
    assert region.bounding_box.height == pytest.approx(4)


def test_Ellipse_visual():
    region = Ellipse((10, 10), 4, 2, 90)
    _fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.plot(*region.points.T, marker="o", color="Blue")
    ax.plot(*np.array(region.shapely_object.exterior.coords).T, marker=".", color="Red")
    ax.add_patch(region.as_artist(origin=(0, 0), fill=True, alpha=0.2))
    ax.plot(*np.array(region.buffer(1).exterior.coords).T, marker=".", color="Yellow")
    ax.plot(*region.centroid, "*", color="Green")
    region.plot(color="Green", alpha=0.2)
    # plt.show()
    plt.close("all")


def test_Polygon():
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    region = Polygon(points)
    assert (10, 1) not in region
    assert (0.5, 0.5) in region
    assert np.array_equal(region.points, points)
    assert (
        repr(region)
        == "Polygon([[0.0, 0.0], [0.0, 1.0], [1.0, 1.0], [1.0, 0.5], [0.0, 0.0]])"
    )
    region = Polygon(points[:-1])
    assert isinstance(region, Region)
    assert (
        repr(region)
        == "Polygon([[0.0, 0.0], [0.0, 1.0], [1.0, 1.0], [1.0, 0.5], [0.0, 0.0]])"
    )
    assert str(region) == "Polygon(<self.points>, <self.holes>)"
    new_reg = eval(repr(region))
    assert isinstance(new_reg, Polygon)
    assert region.dimension == 2
    assert np.array_equal(region.points, points)
    assert np.array_equal(
        region.centroid,
        (
            pytest.approx(0.4444444444444444),
            pytest.approx(0.611111111111111),
        ),
    )
    assert region.max_distance == np.sqrt(2)
    assert region.region_measure == pytest.approx(0.75)
    assert region.subregion_measure == pytest.approx(3.618033988749895)

    assert np.array_equal(
        region.contains([[0, 0], [0.2, 0.8], [100, 100], [1, 0.5]]), (1,)
    )
    assert region.contains([(0.2, 0.8)]) == (0,)
    assert region.contains([(100, 100)]).size == 0
    assert region.contains([]).size == 0
    assert isinstance(region.as_artist(), mPatches.PathPatch)
    assert isinstance(region.shapely_object, shPolygon)
    assert region.region_measure == pytest.approx(region.shapely_object.area, rel=10e-3)
    assert isinstance(region.buffer(1), Polygon)
    assert np.array_equal(region.bounding_box.corner, (0, 0))
    assert region.bounding_box.width == pytest.approx(1)
    assert region.bounding_box.height == pytest.approx(1)


def test_Polygon_visual():
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    region = Polygon(points)
    _fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.plot(*region.points.T, marker=".", color="Blue")
    ax.plot(*np.array(region.shapely_object.exterior.coords).T, marker=".", color="Red")
    ax.add_patch(region.as_artist(fill=True, alpha=0.2))
    ax.plot(*np.array(region.buffer(1).exterior.coords).T, marker=".", color="Yellow")
    ax.plot(*region.centroid, "*", color="Green")
    region.plot(color="Green", alpha=0.2)
    # plt.show()

    # visualize points inside
    # points = np.random.default_rng().random(size=(10, 2))
    # points_inside = points[region.contains(points)]
    # fig, ax = plt.subplots(nrows=1, ncols=1)
    # ax.scatter(*points_inside.T, marker='.', color='Blue')
    # ax.plot(*np.array(region.shapely_object.exterior.coords).T, marker='.', color='Red')
    # ax.add_patch(region.as_artist(fill=True, alpha=0.2))
    # plt.show()

    plt.close("all")


def test_Polygon_with_holes():
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [
        ((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)),
        ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45)),
    ]
    region = Polygon(points, holes)
    assert np.array_equal(region.points, points)
    assert (
        repr(region)
        == "Polygon([[0.0, 0.0], [0.0, 1.0], [1.0, 1.0], [1.0, 0.5], [0.0, 0.0]], "
        "[[[0.2, 0.2], [0.2, 0.3], [0.3, 0.3], [0.3, 0.25]], "
        "[[0.4, 0.4], [0.4, 0.5], [0.5, 0.5], [0.5, 0.45]]])"
    )
    region = Polygon(points[:-1], holes)
    assert isinstance(region, Region)
    assert (
        repr(region)
        == "Polygon([[0.0, 0.0], [0.0, 1.0], [1.0, 1.0], [1.0, 0.5], [0.0, 0.0]], "
        "[[[0.2, 0.2], [0.2, 0.3], [0.3, 0.3], [0.3, 0.25]], "
        "[[0.4, 0.4], [0.4, 0.5], [0.5, 0.5], [0.5, 0.45]]])"
    )
    assert str(region) == "Polygon(<self.points>, <self.holes>)"
    new_reg = eval(repr(region))
    assert isinstance(new_reg, Polygon)
    assert region.dimension == 2
    assert np.array_equal(region.points, points)
    assert np.array_equal(
        region.centroid,
        (
            pytest.approx(0.446485260770975),
            pytest.approx(0.6162131519274376),
        ),
    )
    assert region.max_distance == np.sqrt(2)
    assert region.region_measure == pytest.approx(0.735)
    assert region.subregion_measure == pytest.approx(4.341640786499874)

    assert np.array_equal(
        region.contains([[0, 0], [0.2, 0.8], [100, 100], [1, 0.5]]), (1,)
    )
    assert region.contains([(0.2, 0.8)]) == (0,)
    assert region.contains([(100, 100)]).size == 0
    assert region.contains([]).size == 0
    assert isinstance(region.as_artist(), mPatches.PathPatch)
    assert isinstance(region.shapely_object, shPolygon)
    assert region.region_measure == pytest.approx(region.shapely_object.area, rel=10e-3)
    assert isinstance(region.buffer(1), Polygon)
    assert np.array_equal(region.bounding_box.corner, (0, 0))
    assert region.bounding_box.width == pytest.approx(1)
    assert region.bounding_box.height == pytest.approx(1)


def test_Polygon_with_holes_visual():
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [
        ((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)),
        ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45)),
    ]
    region = Polygon(points, holes)
    _fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.plot(*region.points.T, marker="o", color="Blue")
    ax.plot(*np.array(region.shapely_object.exterior.coords).T, marker=".", color="Red")
    ax.add_patch(region.as_artist(fill=True, alpha=0.2))
    ax.plot(
        *np.array(region.buffer(0.01).exterior.coords).T, marker=".", color="Yellow"
    )
    ax.plot(*region.centroid, "*", color="Green")
    region.plot(color="Green", alpha=0.2)
    # plt.show()

    plt.close("all")


def test_Polygon_from_shapely():
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [
        ((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)),
        ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45)),
    ]
    shapely_polygon = shPolygon(points, holes)
    region = Polygon.from_shapely(shapely_polygon)
    assert np.array_equal(region.points, points)
    assert isinstance(region, Region)
    assert (
        repr(region)
        == "Polygon([[0.0, 0.0], [0.0, 1.0], [1.0, 1.0], [1.0, 0.5], [0.0, 0.0]], "
        "[[[0.2, 0.2], [0.2, 0.3], [0.3, 0.3], [0.3, 0.25], [0.2, 0.2]], "
        "[[0.4, 0.4], [0.4, 0.5], [0.5, 0.5], [0.5, 0.45], [0.4, 0.4]]])"
    )
    assert str(region) == "Polygon(<self.points>, <self.holes>)"
    assert region.dimension == 2
    assert np.array_equal(region.points, points)
    assert np.array_equal(
        region.centroid,
        (
            pytest.approx(0.446485260770975),
            pytest.approx(0.6162131519274376),
        ),
    )
    assert region.max_distance == np.sqrt(2)
    assert region.region_measure == pytest.approx(0.735)
    assert region.subregion_measure == pytest.approx(4.341640786499874)

    assert np.array_equal(
        region.contains([[0, 0], [0.2, 0.8], [100, 100], [1, 0.5]]), (1,)
    )
    assert region.contains([(0.2, 0.8)]) == (0,)
    assert region.contains([(100, 100)]).size == 0
    assert region.contains([]).size == 0
    assert isinstance(region.as_artist(), mPatches.PathPatch)
    assert isinstance(region.shapely_object, shPolygon)
    assert region.region_measure == pytest.approx(region.shapely_object.area, rel=10e-3)
    assert isinstance(region.buffer(1), Polygon)


@pytest.mark.visual
def test_Polygon_from_shapely_visual():
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [
        ((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)),
        ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45)),
    ]
    shapely_polygon = shPolygon(points, holes)
    region = Polygon.from_shapely(shapely_polygon)

    _fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.plot(*region.points.T, marker="o", color="Blue")
    ax.plot(*np.array(region.shapely_object.exterior.coords).T, marker=".", color="Red")
    ax.add_patch(region.as_artist(fill=True, alpha=0.2))
    ax.plot(
        *np.array(region.buffer(0.01).exterior.coords).T, marker=".", color="Yellow"
    )
    ax.plot(*region.centroid, "*", color="Green")
    region.plot(color="Green", alpha=0.2)
    plt.show()

    plt.close("all")


def test_MultiPolygon():
    points = ((2, 2), (2, 3), (3, 3), (3, 2.5), (2, 2))
    region_0 = Polygon(points)
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [
        ((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)),
        ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45)),
    ]
    region_1 = Polygon(points, holes)
    region = MultiPolygon([region_0, region_1])
    assert (10, 1) not in region
    assert (0.2, 0.8) in region
    assert isinstance(region, Region)
    assert (
        repr(region)
        == "MultiPolygon([Polygon([[2.0, 2.0], [2.0, 3.0], [3.0, 3.0], [3.0, 2.5], [2.0, 2.0]]), Polygon([[0.0, 0.0], [0.0, 1.0], [1.0, 1.0], [1.0, 0.5], [0.0, 0.0]], [[[0.2, 0.2], [0.2, 0.3], [0.3, 0.3], [0.3, 0.25]], [[0.4, 0.4], [0.4, 0.5], [0.5, 0.5], [0.5, 0.45]]])])"
    )
    assert str(region) == "MultiPolygon(<self.polygons>)"
    new_reg = eval(repr(region))
    assert isinstance(new_reg, MultiPolygon)
    assert region.dimension == 2
    assert len(region.polygons) == 2
    assert len(region.points) == 2
    assert len(region.holes) == 2
    assert np.array_equal(
        region.centroid,
        (
            pytest.approx(1.4555555555555553),
            pytest.approx(1.6237373737373733),
        ),
    )
    assert region.max_distance == pytest.approx(4.242640687119285)
    assert region.region_measure == pytest.approx(1.485)
    assert region.subregion_measure == pytest.approx(7.959674775249769)
    assert np.array_equal(
        region.contains([[0, 0], [0.2, 0.8], [100, 100], [1, 0.5]]), (1,)
    )
    assert region.contains([(0.2, 0.8)]) == (0,)
    assert region.contains([(100, 100)]).size == 0
    assert region.contains([]).size == 0
    assert isinstance(region.as_artist(), mPatches.PathPatch)
    assert isinstance(region.shapely_object, shMultiPolygon)
    assert region.region_measure == pytest.approx(region.shapely_object.area, rel=10e-3)
    assert isinstance(region.buffer(1), Polygon)
    assert np.array_equal(region.bounds, (0, 0, 3, 3))
    assert np.array_equal(region.extent, (3, 3))
    assert np.array_equal(region.bounding_box.corner, (0, 0))
    assert region.bounding_box.width == pytest.approx(3)
    assert region.bounding_box.height == pytest.approx(3)


def test_MultiPolygon_visual():
    points = ((2, 2), (2, 3), (3, 3), (3, 2.5), (2, 2))
    region_0 = Polygon(points)
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [
        ((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)),
        ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45)),
    ]
    region_1 = Polygon(points, holes)
    region = MultiPolygon([region_0, region_1])

    _fig, ax = plt.subplots(nrows=1, ncols=1)
    for polygon in region.polygons:
        ax.plot(*polygon.points.T, marker="o", color="Blue")
        ax.plot(
            *np.array(polygon.shapely_object.exterior.coords).T, marker=".", color="Red"
        )
    ax.add_patch(region.as_artist(fill=True, alpha=0.2))
    ax.plot(*np.array(region.buffer(1).exterior.coords).T, marker=".", color="Yellow")
    ax.plot(*region.centroid, "*", color="Green")
    region.plot(color="Green", alpha=0.2)
    # plt.show()

    # visualize points inside
    # points = np.random.default_rng().random(size=(100, 2)) * 5
    # points_inside = points[region.contains(points)]
    # fig, ax = plt.subplots(nrows=1, ncols=1)
    # ax.plot(*points.T, '.', markersize=1, color='Gray')
    # ax.plot(*points_inside.T, 'o', markersize=1, color='Blue')
    # for polygon in region.polygons:
    #     ax.plot(*np.array(polygon.shapely_object.exterior.coords).T, marker='.', color='Red')
    # ax.add_patch(region.as_artist(fill=True, alpha=0.2))
    # plt.show()

    plt.close("all")


def test_Polygon_union():
    points = ((2, 2), (2, 3), (3, 3), (3, 2.5), (2, 2))
    region_0 = Polygon(points)
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [
        ((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)),
        ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45)),
    ]
    region_1 = Polygon(points, holes)
    region = region_0.union(region_1)
    assert isinstance(region, Region)
    assert isinstance(region, Region2D)
    assert isinstance(region, MultiPolygon)
    assert region.dimension == 2
    assert len(region.polygons) == 2
    assert len(region.points) == 2
    assert len(region.holes) == 2


@pytest.mark.visual
def test_Polygon_union_visual():
    points = ((2, 2), (2, 3), (3, 3), (3, 2.5), (2, 2))
    region_0 = Polygon(points)
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [
        ((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)),
        ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45)),
    ]
    region_1 = Polygon(points, holes)
    region = region_0.union(region_1)

    _fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.add_patch(region_0.as_artist(fill=True, alpha=0.2, color="Red"))
    ax.add_patch(region_1.as_artist(fill=True, alpha=0.2, color="Green"))
    ax.add_patch(region.as_artist(fill=True, alpha=0.2, color="Blue"))
    ax.plot(*region.centroid, "*", color="Green")
    region.plot(color="Green", alpha=0.2)
    plt.show()

    # visualize points inside
    points = np.random.default_rng().random(size=(1_000, 2)) * 10
    points_inside = points[region.contains(points)]
    _fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.plot(*points_inside.T, ".", markersize=1, color="Blue")
    ax.add_patch(region.as_artist(fill=True, alpha=0.2))
    plt.show()

    plt.close("all")


def test_Polygon_union_ragged():
    points = ((0, 0), (0, 10), (10, 10), (10, 0))
    holes = [((1, 1), (1, 9), (4, 9), (4, 1))]
    region_0 = Polygon(points, holes)
    points = ((2, 2), (2, 5), (3, 5), (3, 2))
    holes = [((2.5, 2.5), (2.5, 4), (2.6, 4), (2.6, 2.5))]
    region_1 = Polygon(points, holes)
    region = region_0.union(region_1)
    assert region.region_measure == pytest.approx(
        region_0.region_measure + region_1.region_measure
    )


@pytest.mark.visual
def test_Polygon_union_ragged_visual():
    points = ((0, 0), (0, 10), (10, 10), (10, 0))
    holes = [((1, 1), (1, 9), (4, 9), (4, 1))]
    region_0 = Polygon(points, holes)
    points = ((2, 2), (2, 5), (3, 5), (3, 2))
    holes = [((2.5, 2.5), (2.5, 4), (2.6, 4), (2.6, 2.5))]
    region_1 = Polygon(points, holes)
    region = region_0.union(region_1)

    _fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.add_patch(region_0.as_artist(fill=True, alpha=0.2, color="Red"))
    ax.add_patch(region_1.as_artist(fill=True, alpha=0.2, color="Green"))
    ax.add_patch(region.as_artist(fill=True, alpha=0.2, color="Blue"))
    ax.plot(*region.centroid, "*", color="Green")
    region.plot(color="Green", alpha=0.2)
    plt.show()

    # visualize points inside
    points = np.random.default_rng().random(size=(1_000, 2)) * 10
    points_inside = points[region.contains(points)]
    _fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.plot(*points_inside.T, ".", markersize=1, color="Blue")
    ax.add_patch(region.as_artist(fill=True, alpha=0.2))
    plt.show()

    plt.close("all")


def test_Polygon_intersection():
    points = ((0, 2), (0, 3), (3, 3), (3, 2.5), (2, 2))
    region_0 = Polygon(points)
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [
        ((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)),
        ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45)),
    ]
    region_1 = Polygon(points, holes)
    region = region_0.intersection(region_1)
    assert isinstance(region, EmptyRegion)

    points = ((0, 0), (0, 3), (3, 3), (3, 2), (2, 1.5))
    region_0 = Polygon(points)
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [
        ((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)),
        ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45)),
    ]
    region_1 = Polygon(points, holes)
    region = region_0.intersection(region_1)
    assert isinstance(region, Polygon)
    assert region.dimension == 2
    assert len(region.points) == 5
    assert len(region.holes) == 2


@pytest.mark.visual
def test_Polygon_intersection_visual():
    points = ((0, 0), (0, 3), (3, 3), (3, 2), (2, 1.5))
    region_0 = Polygon(points)
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [
        ((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)),
        ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45)),
    ]
    region_1 = Polygon(points, holes)
    region = region_0.intersection(region_1)

    _fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.add_patch(region_0.as_artist(fill=True, alpha=0.2, color="Red"))
    ax.add_patch(region_1.as_artist(fill=True, alpha=0.2, color="Green"))
    ax.add_patch(region.as_artist(fill=True, alpha=0.2, color="Blue"))
    ax.plot(*region.centroid, "*", color="Green")
    region.plot(color="Green", alpha=0.2)
    plt.show()

    # visualize points inside
    points = np.random.default_rng().random(size=(100_000, 2)) * 10
    points_inside = points[region.contains(points)]
    _fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.plot(*points_inside.T, ".", markersize=1, color="Blue")
    ax.add_patch(region.as_artist(fill=True, alpha=0.2))
    plt.show()

    plt.close("all")


def test_Polygon_difference():
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [
        ((0.2, 0.2), (0.2, 0.4), (0.4, 0.4), (0.3, 0.25)),
        ((0.5, 0.5), (0.5, 0.8), (0.8, 0.8), (0.7, 0.45)),
    ]
    region_0 = Polygon(points, holes)
    other_region = Rectangle(corner=(0.5, 0.2), width=0.5, height=0.5, angle=0)
    region = region_0.symmetric_difference(other_region)

    points = ((0, 0), (0, 3), (3, 3), (3, 2), (2, 1.5))
    region_0 = Polygon(points)
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [
        ((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)),
        ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45)),
    ]
    other_region = Polygon(points, holes)
    region = region_0.symmetric_difference(other_region)
    # print(repr(region))
    assert isinstance(region, MultiPolygon)
    assert region.dimension == 2
    assert len(region.polygons) == 4


@pytest.mark.visual
def test_Polygon_difference_visual():
    points = ((0, 0), (0, 3), (3, 3), (3, 2), (2, 1.5))
    region_0 = Polygon(points)
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [
        ((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)),
        ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45)),
    ]
    other_region = Polygon(points, holes)
    region = region_0.symmetric_difference(other_region)

    _fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.add_patch(region_0.as_artist(fill=True, alpha=0.2, color="Red"))
    ax.add_patch(other_region.as_artist(fill=True, alpha=0.2, color="Green"))
    ax.add_patch(region.as_artist(fill=True, alpha=0.2, color="Blue"))
    ax.plot(*region.centroid, "*", color="Green")
    region.plot(color="Green", alpha=0.2)
    plt.show()

    # visualize points inside
    points = np.random.default_rng().random(size=(100_000, 2)) * 10
    points_inside = points[region.contains(points)]
    _fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.plot(*points_inside.T, ".", markersize=1, color="Blue")
    ax.add_patch(region.as_artist(fill=True, alpha=0.2))
    plt.show()

    plt.close("all")


def test_AxisOrientedCuboid():
    region = AxisOrientedCuboid((1, 1, 1), 9, 19, 29)
    assert (-1, 1, 1) not in region
    assert (2, 2, 2) in region
    assert isinstance(region, Region)
    assert isinstance(region, Region3D)
    assert repr(region) == "AxisOrientedCuboid((1, 1, 1), 9, 19, 29)"
    assert str(region) == "AxisOrientedCuboid((1, 1, 1), 9, 19, 29)"
    new_reg = eval(repr(region))
    assert isinstance(new_reg, AxisOrientedCuboid)
    with pytest.raises(AttributeError):
        region.corner = None
        region.length = None
        region.width = None
        region.height = None
    assert region.dimension == 3
    assert region.bounds == pytest.approx((1, 1, 1, 10, 20, 30))
    assert np.array_equal(region.intervals, ((1, 10), (1, 20), (1, 30)))
    assert region.extent == pytest.approx((9, 19, 29))
    assert len(region.points) == 8
    assert np.array_equal(region.centroid, (5.5, 10.5, 15.5))
    assert region.max_distance == np.sqrt(9**2 + 19**2 + 29**2)
    assert region.region_measure == (9 * 19 * 29)
    assert region.subregion_measure == 2 * (9 * 19 + 19 * 29 + 29 * 9)
    assert np.array_equal(
        region.contains(
            [[0, 0, 0], [1, 10, 10], [10, 10, 10], [5, 100, 10], [5, 10, 100]]
        ),
        (1,),
    )
    # other = Rectangle((0, 0), 2, 1, 0)
    # assert isinstance(region.intersection(other), Polygon)
    # assert isinstance(region.symmetric_difference(other), MultiPolygon)
    # assert isinstance(region.union(other), Polygon)

    assert region.contains([(2, 2, 2)]) == (0,)
    assert region.contains([(0, 0, 0)]).size == 0
    assert region.contains([]).size == 0
    # assert isinstance(region.as_artist(), mPatches.Rectangle)
    assert isinstance(region.buffer(1), AxisOrientedCuboid)
    assert repr(region.buffer(1)) == "AxisOrientedCuboid((0, 0, 0), 11, 21, 31)"
    assert np.array_equal(region.bounding_box.corner, (1, 1, 1))
    assert region.bounding_box.length == pytest.approx(9)
    assert region.bounding_box.width == pytest.approx(19)
    assert region.bounding_box.height == pytest.approx(29)

    region = AxisOrientedCuboid.from_intervals(((1, 10), (1, 20), (1, 30)))
    assert repr(region) == "AxisOrientedCuboid((1, 1, 1), 9, 19, 29)"
    region = AxisOrientedCuboid.from_intervals([(1, 10), (1, 20), (1, 30)])
    assert repr(region) == "AxisOrientedCuboid((1, 1, 1), 9, 19, 29)"
    region = AxisOrientedCuboid.from_intervals(np.array([(1, 10), (1, 20), (1, 30)]))
    assert repr(region) == "AxisOrientedCuboid((1, 1, 1), 9, 19, 29)"


def test_Cuboid():
    region = Cuboid((1, 1, 1), 9, 19, 29, 0, 0, 0)
    assert isinstance(region, Region)
    assert isinstance(region, Region3D)
    assert repr(region) == "Cuboid((1, 1, 1), 9, 19, 29, 0, 0, 0)"
    assert str(region) == "Cuboid((1, 1, 1), 9, 19, 29, 0, 0, 0)"
    new_reg = eval(repr(region))
    assert isinstance(new_reg, Cuboid)
    with pytest.raises(AttributeError):
        region.corner = None
        region.length = None
        region.width = None
        region.height = None
        region.alpha = None
        region.beta = None
        region.gamma = None
    assert region.dimension == 3
    # assert region.bounds == pytest.approx((1, 1, 1, 10, 20, 30))
    # assert region.extent == pytest.approx((9, 19, 29))
    # assert len(region.points) == 8
    # assert region.centroid == (5.5, 10.5, 15.5)
    assert region.max_distance == np.sqrt(9**2 + 19**2 + 29**2)
    assert region.region_measure == (9 * 19 * 29)
    assert region.subregion_measure == 2 * (9 * 19 + 19 * 29 + 29 * 9)
    # assert np.array_equal(region.contains([[0, 0, 0], [1, 10, 10], [10, 10, 10], [5, 100, 10], [5, 10, 100]]), (1,))
    # other = Rectangle((0, 0), 2, 1, 0)
    # assert isinstance(region.intersection(other), Polygon)
    # assert isinstance(region.symmetric_difference(other), MultiPolygon)
    # assert isinstance(region.union(other), Polygon)

    # assert region.contains([(2, 2, 2)]) == (0,)
    # assert region.contains([(0, 0, 0)]).size == 0
    # assert region.contains([]).size == 0
    # assert isinstance(region.as_artist(), mPatches.Rectangle)
    # assert isinstance(region.buffer(1), AxisOrientedCuboid)
    # assert repr(region.buffer(1)) == 'AxisOrientedCuboid((0, 0, 0), 11, 21, 31)'
    # assert region.bounding_box.corner == (1, 1, 1)
    # assert region.bounding_box.length == pytest.approx(9)
    # assert region.bounding_box.width == pytest.approx(19)
    # assert region.bounding_box.height == pytest.approx(29)


def test_AxisOrientedHypercuboid():
    region = AxisOrientedHypercuboid((1, 1, 1), (9, 19, 29))
    assert (-1, 1, 1) not in region
    assert (2, 2, 2) in region
    assert isinstance(region, Region)
    assert repr(region) == "AxisOrientedHypercuboid((1, 1, 1), (9, 19, 29))"
    assert str(region) == "AxisOrientedHypercuboid((1, 1, 1), (9, 19, 29))"
    new_reg = eval(repr(region))
    assert isinstance(new_reg, AxisOrientedHypercuboid)
    with pytest.raises(AttributeError):
        region.corner = None
        region.length = None
        region.width = None
        region.height = None
    assert region.dimension == 3
    assert region.bounds == pytest.approx((1, 1, 1, 10, 20, 30))
    assert np.array_equal(region.intervals, ((1, 10), (1, 20), (1, 30)))
    assert region.extent == pytest.approx((9, 19, 29))
    assert len(region.points) == 8
    assert region.centroid.tolist() == [5.5, 10.5, 15.5]
    assert region.max_distance == np.sqrt(9**2 + 19**2 + 29**2)
    assert region.region_measure == (9 * 19 * 29)
    with pytest.raises(NotImplementedError):
        region.subregion_measure  # noqa B018
    assert np.array_equal(
        region.contains(
            [[0, 0, 0], [1, 10, 10], [10, 10, 10], [5, 100, 10], [5, 10, 100]]
        ),
        (1,),
    )
    # needs to be implemented:
    # other = Rectangle((0, 0), 2, 1, 0)
    # assert isinstance(region.intersection(other), Polygon)
    # assert isinstance(region.symmetric_difference(other), MultiPolygon)
    # assert isinstance(region.union(other), Polygon)

    assert region.contains([(2, 2, 2)]) == (0,)
    assert region.contains([(0, 0, 0)]).size == 0
    assert region.contains([]).size == 0
    # assert isinstance(region.as_artist(), mPatches.Rectangle)
    assert isinstance(region.buffer(1), AxisOrientedHypercuboid)
    assert repr(region.buffer(1)) == "AxisOrientedHypercuboid((0, 0, 0), (11, 21, 31))"
    assert region.bounding_box.corner.tolist() == [1, 1, 1]
    assert region.bounding_box.lengths.tolist() == [9, 19, 29]

    region = AxisOrientedHypercuboid.from_intervals(((1, 10), (1, 20), (1, 30)))
    assert repr(region) == "AxisOrientedHypercuboid((1, 1, 1), (9, 19, 29))"
    region = AxisOrientedHypercuboid.from_intervals([(1, 10), (1, 20), (1, 30)])
    assert repr(region) == "AxisOrientedHypercuboid((1, 1, 1), (9, 19, 29))"
    region = AxisOrientedHypercuboid.from_intervals(
        np.array([(1, 10), (1, 20), (1, 30)])
    )
    assert repr(region) == "AxisOrientedHypercuboid((1, 1, 1), (9, 19, 29))"


def test_pickling_Region():
    regions = [
        EmptyRegion(),
        Interval(),
        Rectangle(),
        Ellipse(),
        Polygon(((0, 0), (0, 1), (2, 2))),
        MultiPolygon(
            [Polygon(((10, 10), (10, 11), (12, 12))), Polygon(((0, 0), (0, 1), (2, 2)))]
        ),
    ]
    with tempfile.TemporaryDirectory() as tmp_directory:
        file_path = Path(tmp_directory) / "pickled_region.pickle"
        for region in regions:
            with open(file_path, "wb") as file:
                pickle.dump(region, file, pickle.HIGHEST_PROTOCOL)
            with open(file_path, "rb") as file:
                region_pickled = pickle.load(file)  # noqa S301
            assert type(region_pickled) is type(region)


def test__polygon_path():
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [
        ((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)),
        ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45)),
    ]
    shapely_polygon = shPolygon(points, holes)
    path = _polygon_path(polygon=shapely_polygon)
    assert isinstance(path, mplPath)

    polygon = Polygon(((0, 0), (0, 1), (2, 2)))
    path = _polygon_path(polygon=polygon)
    assert isinstance(path, mplPath)



"""
Benchmark functions for :mod:`locan.analysis.ripley`
"""

import numpy as np

import locan as lc
from locan.analysis.ripley import (
    _ripleys_k_function,
)


class BenchmarkRipley:
    """
    Benchmarks for Ripley's k function computations
    """

    def setup(self):
        path = lc.ROOT_DIR / "tests/test_data/five_blobs.txt"
        self.locdata = lc.load_txt_file(path)
        print(self.locdata.data.info())
        self.radii = np.arange(10, 110, 10)
        print(self.radii)

    def check__ripleys_k_function(self):
        results_df = _ripleys_k_function(
            points=self.locdata.coordinates, radii=self.radii
        )
        print(results_df)

    def time__ripleys_k_function(self):
        _ripleys_k_function(points=self.locdata.coordinates, radii=self.radii)


def main():
    bm = BenchmarkRipley()
    bm.setup()
    bm.check__ripleys_k_function()


def main_profile():
    """
    Run benchmarks through profiler.
    """
    import cProfile
    import pstats

    bm = BenchmarkRipley()
    bm.setup()

    profiler = cProfile.Profile()
    profiler.enable()

    for _ in range(100):
        bm.time__ripleys_k_function()

    profiler.disable()
    stats = pstats.Stats(profiler).sort_stats("cumtime")
    stats.print_stats("time_")


if __name__ == "__main__":
    main()
    main_profile()



"""

Render localization data.

This module provides convenience functions for rendering localization data.

"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from locan.configuration import RENDER_ENGINE
from locan.constants import RenderEngine
from locan.dependencies import HAS_DEPENDENCY
from locan.visualize.render_mpl.render2d import render_2d_mpl, render_2d_scatter_density
from locan.visualize.render_napari.render2d import render_2d_napari
from locan.visualize.render_napari.render3d import render_3d_napari

if TYPE_CHECKING:
    from locan.data.locdata import LocData

__all__: list[str] = [
    "render_2d",
    "render_3d",
]

logger = logging.getLogger(__name__)


def render_2d(
    locdata: LocData, render_engine: RenderEngine = RENDER_ENGINE, **kwargs: Any
) -> Any:
    """
    Wrapper function to render localization data into a 2D image.
    For complete signatures see render_2d_mpl or corresponding functions.
    """
    if render_engine == RenderEngine.MPL:
        return render_2d_mpl(locdata, **kwargs)
    elif (
        HAS_DEPENDENCY["mpl_scatter_density"]
        and render_engine == RenderEngine.MPL_SCATTER_DENSITY
    ):
        return render_2d_scatter_density(locdata, **kwargs)
    elif HAS_DEPENDENCY["napari"] and render_engine == RenderEngine.NAPARI:
        return render_2d_napari(locdata, **kwargs)
    else:
        raise NotImplementedError(f"render_2d is not implemented for {render_engine}.")


def render_3d(
    locdata: LocData, render_engine: RenderEngine = RENDER_ENGINE, **kwargs: Any
) -> Any:
    """
    Wrapper function to render localization data into a 3D image.
    For complete signatures see render_3d_mpl or corresponding functions.
    """
    if HAS_DEPENDENCY["napari"] and render_engine == RenderEngine.NAPARI:
        return render_3d_napari(locdata, **kwargs)
    else:
        raise NotImplementedError(f"render_3d is not implemented for {render_engine}.")



"""

Aggregate localization data.

This module provides functions to bin LocData objects to form a histogram or
image.

Specify bins through one of the parameters (`bins`, `bin_edges`, `n_bins`,
`bin_size`, `bin_range`, `labels`) as further outlined in the documentation
for :class:`Bins`.
"""

from __future__ import annotations

import logging
import warnings
from collections import namedtuple
from collections.abc import Iterable, Sequence
from math import isclose
from typing import Any, Literal, cast

import boost_histogram as bh
import fast_histogram
import numpy as np
import numpy.typing as npt

from locan.data.locdata import LocData
from locan.data.validation import _check_loc_properties
from locan.process.properties.locdata_statistics import ranges

__all__: list[str] = ["Bins", "histogram"]

logger = logging.getLogger(__name__)


def is_array_like(anything: Any) -> bool:
    """
    Return true if `anything` can be turned into a numpy.ndarray without
    creating elements of type object.

    Catches numpy.exceptions.VisibleDeprecationWarning or ValueError when setting an
    array element with a sequence.

    Parameters
    ----------
    anything
        Anything to be classified as being array-like or not.

    Returns
    -------
    bool
    """
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        try:
            return np.asarray(anything).dtype != object  # noqa: E721
        except np.exceptions.VisibleDeprecationWarning as e:
            if "Creating an ndarray from ragged nested sequences" in str(e):
                return False
            else:
                raise e
        except ValueError as e:
            if "setting an array element with a sequence." in str(e):
                return False
            else:
                raise e
        except Exception as e:
            raise e


def _is_scalar(element: Any) -> bool:
    return isinstance(element, (int, float)) or (
        is_array_like(element) and np.size(element) == 1 and np.ndim(element) == 0
    )


def _is_single_element(element: Any) -> bool:
    return isinstance(element, (int, float)) or (
        is_array_like(element) and np.size(element) == 1 and np.ndim(element) in (0, 1)
    )


def _is_1d_array_of_scalar(element: Any) -> bool:
    return is_array_like(element) and np.size(element) >= 1 and np.ndim(element) == 1


def _is_1d_array_of_two_or_more_scalar(element: Any) -> bool:
    return is_array_like(element) and np.size(element) > 1 and np.ndim(element) == 1


def _is_1d_array_of_two_scalar(element: Any) -> bool:
    return is_array_like(element) and np.size(element) == 2 and np.ndim(element) == 1


def _is_2d_homogeneous_array(element: Any) -> bool:
    return is_array_like(element) and np.size(element) > 1 and np.ndim(element) == 2


def _is_2d_inhomogeneous_array_of_1d_array_of_scalar(element: Any) -> bool:
    return not is_array_like(element) and all(
        _is_1d_array_of_scalar(el) for el in element
    )


def _is_2d_inhomogeneous_array(element: Any) -> bool:
    return not is_array_like(element) and all(
        _is_scalar(el) or _is_1d_array_of_scalar(el) for el in element
    )


def _is_2d_array_of_1d_array_of_scalar(element: Any) -> bool:
    return _is_2d_homogeneous_array(
        element
    ) or _is_2d_inhomogeneous_array_of_1d_array_of_scalar(element)


def _n_bins_to_bin_edges_one_dimension(
    n_bins: int, bin_range: tuple[float, float] | Sequence[float]
) -> npt.NDArray[np.float64]:
    """
    Compute bin edges from n_bins and bin_range.

    Parameters
    ----------
    n_bins
        Number of bins.
    bin_range
        Minimum and maximum edge. Array with shape (2,).

    Returns
    -------
    npt.NDArray[np.float64]
        Array with bin edges.
    """
    return np.linspace(*bin_range, n_bins + 1, endpoint=True, dtype=float)  # type: ignore


def _bin_size_to_bin_edges_one_dimension(
    bin_size: float | Sequence[float],
    bin_range: tuple[float, float] | Sequence[float],
    extend_range: bool | None = None,
) -> npt.NDArray[np.float64]:
    """
    Compute bin edges from bin_size and bin_range.

    Use bin_edges if you need to use variable bin_sizes for Bins construction.

    Parameters
    ----------
    bin_size
        One size or sequence of sizes for bins with shape (n_bins,).
    bin_range
        Minimum and maximum edge. Array with shape (2,).
    extend_range
        If for equally-sized bins the final bin_edge is different from the
        maximum bin_range, the last bin_edge will be smaller than the maximum
        bin_range but all bins are equally-sized (None);
        the last bin_edge will be equal to the maximum bin_range but bins
        are not equally-sized (False);
        the last bin_edge will be larger than the maximum bin_range but all
        bins are equally-sized (True).

        If for variable-sized bins the final bin_edge is different from the
        maximum bin_range, the last bin_edge will be smaller than the maximum
        bin_range (None);
        the last bin_edge will be equal to the maximum bin_range (False);
        the last bin_edge will be larger than the maximum bin_range but taken
        from the input sequence (True).

    Returns
    -------
    npt.NDArray[np.float64]
        Array of bin edges
    """
    if _is_scalar(bin_size):
        bin_edges = np.arange(*bin_range, bin_size, dtype=float)  # type: ignore
        if (
            bin_edges.size == 1
        ):  # this is the case if bin_size is greater than the bin_range
            if extend_range is True:
                bin_edges = np.append(bin_edges, bin_edges + bin_size)
            elif extend_range is None or extend_range is False:
                bin_edges = np.array(bin_range)
            else:
                raise ValueError("`extend_range` must be None, True or False.")
        else:
            last_edge = bin_edges[-1] + bin_size
            if isclose(last_edge, bin_range[-1]):
                bin_edges = np.append(bin_edges, last_edge)
            else:
                if extend_range is None:
                    pass
                elif extend_range is True:
                    bin_edges = np.append(bin_edges, last_edge)
                elif extend_range is False:
                    bin_edges = np.append(bin_edges, bin_range[-1])
                else:
                    raise ValueError("`extend_range` must be None, True or False.")

    elif _is_1d_array_of_scalar(bin_size):
        bin_size = cast(Sequence[float], bin_size)
        if bin_size[0] > np.diff(bin_range):
            if extend_range is True:
                bin_edges = np.array([bin_range[0], bin_range[0] + bin_size[0]])
            elif extend_range is None or extend_range is False:
                bin_edges = bin_range
            else:
                raise ValueError("`extend_range` must be None, True or False.")
        else:
            bin_edges_ = np.concatenate(
                (np.asarray([bin_range[0]]), np.cumsum(bin_size) + bin_range[0])
            )
            bin_edges = bin_edges_[bin_edges_ <= bin_range[-1]]
            if extend_range is None:
                pass
            elif extend_range is True:
                if len(bin_edges_) > len(bin_edges):
                    last_edge = bin_edges_[len(bin_edges)]
                    bin_edges = np.append(bin_edges, last_edge)
                else:
                    pass
            elif extend_range is False:
                bin_edges = np.append(bin_edges, bin_range[-1])
            else:
                raise ValueError("`extend_range` must be None, True or False.")

    else:
        raise TypeError("`bin_size` must be 0- or 1-dimensional.")

    return np.array(bin_edges)


def _bin_edges_to_n_bins_one_dimension(bin_edges: Sequence[float]) -> int:
    """
    Return the number of bins.

    Parameters
    ----------
    bin_edges
        Array of bin edges
        with shape (n_bin_edges,).

    Returns
    -------
    int
        Number of bins
    """
    n_bins = len(bin_edges) - 1
    return n_bins


def _bin_edges_to_n_bins(
    bin_edges: Sequence[float] | Sequence[Sequence[float]],
) -> tuple[int, ...]:
    """
    Check if bins are equally sized and return the number of bins.

    Parameters
    ----------
    bin_edges
        Bin edges for each dimension with shape (dimension, n_bin_edges).

    Returns
    -------
    tuple[int, ...]
        Number of bins
    """
    n_bins: tuple[int, ...]
    if _is_1d_array_of_scalar(bin_edges):
        bin_edges = cast("Sequence[float]", bin_edges)
        n_bins = (_bin_edges_to_n_bins_one_dimension(bin_edges),)
    elif _is_2d_array_of_1d_array_of_scalar(bin_edges):
        bin_edges = cast("Sequence[Sequence[float]]", bin_edges)
        n_bins = tuple(_bin_edges_to_n_bins_one_dimension(edges) for edges in bin_edges)
    else:
        raise TypeError("The shape of bin_edges must be (dimension, n_bin_edges).")
    return n_bins


def _bin_edges_to_bin_size_one_dimension(
    bin_edges: Sequence[float],
) -> float | npt.NDArray[np.float64]:
    """
    Compute the sizes of bins.

    Parameters
    ----------
    bin_edges
        Array of bin edges
        with shape (n_bin_edges,).

    Returns
    -------
    float | npt.NDArray[np.float64]
        Bin size for all bins or for each bin.
    """
    differences = np.diff(bin_edges)
    if np.all(differences == differences[0]):
        bin_size: float | npt.NDArray[np.float64] = differences[0]
    elif np.all(np.isclose(differences, differences[0], atol=0)):
        bin_size = differences[0]
        logger.debug(
            "bin_sizes differ by floating point instability with less than rtol=1.e-5"
        )
    else:
        bin_size = differences
    return bin_size


def _bin_edges_to_bin_size(
    bin_edges: Sequence[float] | Sequence[Sequence[float]],
) -> tuple[float, ...] | tuple[npt.NDArray[np.float64], ...]:
    """
    Compute the sizes of bins.

    Parameters
    ----------
    bin_edges
        Bin edges for each dimension with shape (dimension, n_bin_edges).

    Returns
    -------
    tuple[float, ...] | tuple[npt.NDArray[np.float64], ...]
        Bin size for all bins or for each bin in each dimension.
    """
    if _is_1d_array_of_scalar(bin_edges):
        bin_edges = cast("Sequence[float]", bin_edges)
        bin_size = (_bin_edges_to_bin_size_one_dimension(bin_edges),)
    elif _is_2d_array_of_1d_array_of_scalar(bin_edges):
        bin_edges = cast("Sequence[Sequence[float]]", bin_edges)
        bin_size = tuple(  # type: ignore
            _bin_edges_to_bin_size_one_dimension(edges) for edges in bin_edges
        )
    else:
        raise TypeError("The shape of bin_edges must be (dimension, n_bin_edges).")
    return bin_size  # type: ignore


def _bin_edges_to_bin_centers(
    bin_edges: Sequence[float] | Sequence[Sequence[float]],
) -> tuple[npt.NDArray[np.float64], ...]:
    """
    Compute bin centers.

    Parameters
    ----------
    bin_edges
        Bin edges for each dimension with shape (dimension, n_bin_edges).

    Returns
    -------
    tuple[npt.NDArray[np.float64], ...]
        Array of bin centers for each dimension with shape (n_bins,)
    """
    if _is_1d_array_of_scalar(bin_edges):
        bin_edges = cast("Sequence[float]", bin_edges)
        bin_centers = (np.diff(bin_edges) / 2 + bin_edges[0:-1],)
    elif _is_2d_array_of_1d_array_of_scalar(bin_edges):
        bin_edges = cast("Sequence[Sequence[float]]", bin_edges)
        bin_centers = tuple(np.diff(edges) / 2 + edges[0:-1] for edges in bin_edges)  # type: ignore
    else:
        raise TypeError("The shape of bin_edges must be (dimension, n_bin_edges).")
    return bin_centers


def _indices_to_bin_centers(
    bin_edges: Sequence[float] | Sequence[Sequence[float]], indices: npt.ArrayLike
) -> npt.NDArray[np.float64]:
    """
    Compute bin centers for given indices.

    Parameters
    ----------
    bin_edges
        Bin edges for each dimension with shape (dimension, n_bin_edges).
    indices
        Array of multi-dimensional indices with shape (n_indices, dimension),
        e.g. reflecting a list of vertices.

    Returns
    -------
    npt.NDArray[np.float64]
        Selected bin centers with shape (n_indices, dimension)
    """
    bin_centers = _bin_edges_to_bin_centers(bin_edges)
    if len(bin_centers) == 1:
        bin_centers = bin_centers[0]  # type: ignore
    indices = np.asarray(indices)

    if _is_1d_array_of_scalar(bin_centers):
        if _is_scalar(indices):
            selected_bin_centers = bin_centers[indices]
        elif _is_1d_array_of_scalar(indices):
            selected_bin_centers = bin_centers[indices]
        elif _is_2d_homogeneous_array(indices):
            selected_bin_centers = bin_centers[indices]
        else:
            raise TypeError("The given array shapes cannot be processed.")

    elif _is_2d_array_of_1d_array_of_scalar(bin_centers):
        if _is_scalar(indices):
            selected_bin_centers = np.array([bc[indices] for bc in bin_centers])
        elif _is_1d_array_of_scalar(indices):
            selected_bin_centers = np.array([bc[indices] for bc in bin_centers])
        elif _is_2d_homogeneous_array(indices):
            if len(bin_centers) != len(indices.T):
                raise TypeError(
                    "`bin_centers` and `indices` must be for the same dimension."
                )
            else:
                selected_bin_centers = np.array(
                    [bc[idx] for bc, idx in zip(bin_centers, indices.T)]
                ).T
        else:
            raise TypeError("The given array shapes cannot be processed.")
    else:
        raise TypeError("The given array shapes cannot be processed.")

    return selected_bin_centers


class _BinsFromBoostHistogramAxis:
    """
    Adapter class for dealing with `boost-histogram.axis` elements through the
    `bins` parameter in :class:`Bins`.

    Parameters
    ----------
    bins : boost_histogram.axis.Axis | boost_histogram.axis.AxesTuple
    """

    def __init__(self, bins: bh.axis.Axis | bh.axis.AxesTuple) -> None:
        self.dimension: int
        self.bin_range: tuple[tuple[float, float], ...]
        self.bin_edges: tuple[npt.NDArray[np.float64], ...]
        self.n_bins: tuple[int, ...]
        self.bin_size: tuple[float, ...] | tuple[npt.NDArray[np.float64], ...]
        self.bin_centers: tuple[npt.NDArray[np.float64], ...]
        self._bins: bh.axis.Axis | bh.axis.AxesTuple

        if isinstance(bins, bh.axis.Axis):
            self._bins = bins
            self.dimension = 1
            self.n_bins = (self._bins.size,)
            self.bin_size = (np.asarray(self._bins.widths),)
            self.bin_edges = (self._bins.edges,)
            self.bin_range = ((self._bins.edges[0], self._bins.edges[-1]),)
            self.bin_centers = (self._bins.centers,)

        elif isinstance(bins, bh.axis.AxesTuple):
            self._bins = bins
            self.dimension = len(self._bins)
            self.n_bins = self._bins.size
            self.bin_size = tuple(
                np.asarray(arr) for arr in self._bins.widths.flatten()
            )
            self.bin_edges = self._bins.edges.flatten()
            self.bin_range = tuple(
                (axis.edges[0], axis.edges[-1]) for axis in self._bins
            )
            self.bin_centers = self._bins.centers.flatten()

        else:
            raise TypeError

    def __getattr__(self, attr: str) -> Any:
        if attr.startswith("__") and attr.endswith(
            "__"
        ):  # this is needed to enable pickling
            raise AttributeError
        return getattr(self._bins, attr)


class _BinsFromEdges:
    """
    Builder for :class:`Bins`.

    Parameters
    ----------
    bin_edges : Sequence[float] | Sequence[Sequence[float]]
        Bin edges for all or each dimension with shape (dimension, n_bin_edges).
    """

    def __init__(self, bin_edges: Sequence[float] | Sequence[Sequence[float]]) -> None:
        self.dimension: int
        self.bin_range: tuple[tuple[float, float], ...]
        self.bin_edges: tuple[npt.NDArray[np.float64], ...]
        self.n_bins: tuple[int, ...]
        self.bin_size: tuple[float, ...] | tuple[npt.NDArray[np.float64], ...]

        if _is_1d_array_of_scalar(bin_edges):
            bin_edges = cast(Sequence[float], bin_edges)
            self.bin_edges = (np.array(bin_edges),)
            self.dimension = 1
            self.bin_range = ((bin_edges[0], bin_edges[-1]),)
        elif _is_2d_array_of_1d_array_of_scalar(bin_edges):
            bin_edges = cast(Sequence[Sequence[float]], bin_edges)
            self.bin_edges = tuple(np.array(edges) for edges in bin_edges)
            self.dimension = len(self.bin_edges)
            self.bin_range = tuple((edges[0], edges[-1]) for edges in self.bin_edges)
        else:
            raise TypeError("`bin_edges` must have 1 or 2 dimensions.")

        self.n_bins = _bin_edges_to_n_bins(self.bin_edges)  # type: ignore
        self.bin_size = _bin_edges_to_bin_size(self.bin_edges)  # type: ignore


class _BinsFromNumber:
    """
    Builder for :class:`Bins`.

    Parameters
    ----------
    n_bins : int | Sequence[int]
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_range : tuple[float, float] | Sequence[float] | Sequence[Sequence[float]]
        Minimum and maximum edge for all or each dimensions
        with shape (2,) or (dimension, 2).
    """

    def __init__(
        self,
        n_bins: int | Sequence[int],
        bin_range: tuple[float, float] | Sequence[float] | Sequence[Sequence[float]],
    ) -> None:
        self.dimension: int
        self.bin_range: tuple[tuple[float, float], ...]
        self.bin_edges: tuple[npt.NDArray[np.float64], ...]
        self.n_bins: tuple[int, ...]
        self.bin_size: tuple[float, ...] | tuple[npt.NDArray[np.float64], ...]

        if not is_array_like(n_bins) or np.ndim(n_bins) > 1:
            raise TypeError("`n_bins` must be 0- or 1-dimensional.")

        elif _is_scalar(n_bins):
            n_bins = cast("int", n_bins)
            if _is_1d_array_of_two_scalar(bin_range):
                bin_range = cast("tuple[float, float]", bin_range)
                self.dimension = 1
                self.bin_edges = (
                    _n_bins_to_bin_edges_one_dimension(n_bins, bin_range),
                )
                self.n_bins = (n_bins,)
                self.bin_range = (bin_range,)

            elif _is_2d_homogeneous_array(bin_range):
                bin_range = cast("Sequence[tuple[float, float]]", bin_range)
                self.dimension = len(bin_range)
                self.bin_edges = tuple(
                    _n_bins_to_bin_edges_one_dimension(n_bins, single_range)
                    for single_range in bin_range
                )
                self.n_bins = tuple(n_bins for _ in bin_range)
                self.bin_range = tuple(bin_range)

            else:
                raise TypeError("n_bins and/or bin_range have incorrect shapes.")

        elif _is_1d_array_of_scalar(n_bins):
            n_bins = cast("Sequence[int]", n_bins)
            if _is_1d_array_of_two_or_more_scalar(bin_range):
                bin_range = cast("Sequence[float]", bin_range)
                self.dimension = len(n_bins)
                self.bin_edges = tuple(
                    _n_bins_to_bin_edges_one_dimension(n_bins=n, bin_range=bin_range)
                    for n in n_bins
                )
                self.n_bins = tuple(n_bins)
                self.bin_range = tuple(tuple(bin_range) for _ in n_bins)  # type: ignore

            elif _is_2d_homogeneous_array(bin_range):
                bin_range = cast("Sequence[tuple[float, float]]", bin_range)
                if len(n_bins) != len(bin_range):
                    raise TypeError("n_bins and bin_range have incompatible shapes.")
                else:
                    self.dimension = len(n_bins)
                    self.bin_edges = tuple(
                        _n_bins_to_bin_edges_one_dimension(n_bins=b, bin_range=r)
                        for b, r in zip(n_bins, bin_range)
                    )
                    self.n_bins = tuple(n_bins)
                    self.bin_range = tuple(bin_range)

            else:
                raise TypeError("n_bins and/or bin_range have incorrect shapes.")

        else:
            raise TypeError("n_bins and/or bin_range have incorrect shapes.")

        self.bin_size = _bin_edges_to_bin_size(self.bin_edges)  # type: ignore


class _BinsFromSize:
    """
    Builder for :class:`Bins`.

    Parameters
    ----------
    bin_size : float | Sequence[float] | Sequence[Sequence[float]]
        The size of bins for all or each bin and for all or each dimension
        with shape (dimension,) or (dimension, n_bins).
        5 would describe bin_size of 5 for all bins in all dimensions.
        ((2, 5),) yield bins of size (2, 5) for one dimension.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other
        dimension.
        ((2, 5), (1, 3)) yields bins of size (2, 5) for one dimension and
        (1, 3) for the other dimension.
    bin_range : tuple[float, float] | Sequence[float] | Sequence[Sequence[float]]
        Minimum and maximum edge for all or each dimensions
        with shape (2,) or (dimension, 2).
    extend_range : bool | None
        If for equally-sized bins the final bin_edge is different from the
        maximum bin_range, the last bin_edge will be smaller than the maximum
        bin_range but all bins are equally-sized (None);
        the last bin_edge will be equal to the maximum bin_range but bins
        are not equally-sized (False);
        the last bin_edge will be larger than the maximum bin_range but all
        bins are equally-sized (True).

        If for variable-sized bins the final bin_edge is different from the
        maximum bin_range, the last bin_edge will be smaller than the maximum
        bin_range (None);
        the last bin_edge will be equal to the maximum bin_range (False);
        the last bin_edge will be larger than the maximum bin_range but taken
        from the input sequence (True).
    """

    def __init__(
        self,
        bin_size: float | Sequence[float] | Sequence[Sequence[float]],
        bin_range: tuple[float, float] | Sequence[float] | Sequence[Sequence[float]],
        extend_range: bool | None = None,
    ) -> None:
        self.dimension: int
        self.bin_range: tuple[tuple[float, float], ...]
        self.bin_edges: tuple[npt.NDArray[np.float64], ...]
        self.n_bins: tuple[int, ...]
        self.bin_size: tuple[float, ...] | tuple[npt.NDArray[np.float64], ...]

        if _is_scalar(bin_size):
            bin_size = cast("int | float", bin_size)
            if _is_1d_array_of_two_scalar(bin_range):
                bin_range = cast("tuple[float, float] | Sequence[float]", bin_range)
                self.dimension = 1
                self.bin_edges = (
                    _bin_size_to_bin_edges_one_dimension(
                        bin_size, bin_range, extend_range
                    ),
                )
                self.bin_range = (tuple(bin_range),)  # type: ignore

            elif _is_2d_homogeneous_array(bin_range):
                bin_range = cast("Sequence[Sequence[float]]", bin_range)
                self.dimension = len(bin_range)
                self.bin_edges = tuple(
                    _bin_size_to_bin_edges_one_dimension(
                        bin_size, single_range, extend_range
                    )
                    for single_range in bin_range
                )
                self.bin_range = tuple(
                    (edges[0], edges[-1]) for edges in self.bin_edges
                )

            else:
                raise TypeError("bin_size and/or bin_range have incorrect shapes.")

        elif _is_single_element(bin_size):
            if _is_1d_array_of_two_scalar(bin_range):
                bin_range = cast("tuple[float, float] | Sequence[float]", bin_range)
                self.dimension = 1
                self.bin_edges = (
                    _bin_size_to_bin_edges_one_dimension(
                        bin_size[0], bin_range, extend_range  # type: ignore
                    ),
                )
                self.bin_range = (tuple(bin_range),)  # type: ignore

            elif _is_2d_homogeneous_array(bin_range):
                bin_range = cast("Sequence[Sequence[float]]", bin_range)
                self.dimension = len(bin_range)
                self.bin_edges = tuple(
                    _bin_size_to_bin_edges_one_dimension(
                        bin_size[0], single_range, extend_range  # type: ignore
                    )
                    for single_range in bin_range
                )
                self.bin_range = tuple(
                    (edges[0], edges[-1]) for edges in self.bin_edges
                )

            else:
                raise TypeError("bin_size and/or bin_range have incorrect shapes.")

        elif _is_1d_array_of_scalar(bin_size):
            bin_size = cast("Sequence[int | float]", bin_size)
            if _is_1d_array_of_two_scalar(bin_range):
                bin_range = cast("tuple[float, float] | Sequence[float]", bin_range)
                self.dimension = len(bin_size)
                self.bin_edges = tuple(
                    _bin_size_to_bin_edges_one_dimension(
                        bin_size=bs, bin_range=bin_range, extend_range=extend_range
                    )
                    for bs in bin_size
                )
                self.bin_range = tuple(
                    (edges[0], edges[-1]) for edges in self.bin_edges
                )

            elif _is_2d_homogeneous_array(bin_range):
                bin_range = cast("Sequence[Sequence[float]]", bin_range)
                if len(bin_size) != len(bin_range):
                    raise TypeError("bin_size and bin_range have incompatible shapes.")
                else:
                    self.dimension = len(bin_size)
                    self.bin_edges = tuple(
                        _bin_size_to_bin_edges_one_dimension(
                            bin_size=bs, bin_range=br, extend_range=extend_range
                        )
                        for bs, br in zip(bin_size, bin_range)
                    )
                    self.bin_range = tuple(
                        (edges[0], edges[-1]) for edges in self.bin_edges
                    )

            else:
                raise TypeError("bin_size and/or bin_range have incorrect shapes.")

        elif _is_2d_inhomogeneous_array(bin_size) or _is_2d_homogeneous_array(
            bin_size
        ):  # _is_2d_array_of_1d_array_of_scalar(bin_size):
            bin_size = cast("Sequence[Sequence[int | float]]", bin_size)
            if _is_1d_array_of_two_scalar(bin_range):
                bin_range = cast("tuple[float, float] | Sequence[float]", bin_range)
                self.dimension = len(bin_size)
                self.bin_edges = tuple(
                    _bin_size_to_bin_edges_one_dimension(
                        bin_size=bs, bin_range=bin_range, extend_range=extend_range
                    )
                    for bs in bin_size
                )
                self.bin_range = tuple(
                    (edges[0], edges[-1]) for edges in self.bin_edges
                )

            elif _is_2d_homogeneous_array(bin_range):
                bin_range = cast("Sequence[Sequence[float]]", bin_range)
                if len(bin_size) != len(bin_range):
                    raise TypeError("bin_size and bin_range have incompatible shapes.")
                else:
                    self.dimension = len(bin_size)
                    self.bin_edges = tuple(
                        _bin_size_to_bin_edges_one_dimension(
                            bin_size=bs, bin_range=br, extend_range=extend_range
                        )
                        for bs, br in zip(bin_size, bin_range)
                    )
                    self.bin_range = tuple(
                        (edges[0], edges[-1]) for edges in self.bin_edges
                    )

        elif np.ndim(bin_size) > 2:
            raise TypeError(
                "`bin_size` must be 0-, 1- or 2-dimensional. \
                            Construct from bin_edges if you need variable bin_sizes in one dimension."
            )

        else:
            raise TypeError("bin_size and/or bin_range have incorrect shapes.")

        self.n_bins = _bin_edges_to_n_bins(self.bin_edges)  # type: ignore
        self.bin_size = _bin_edges_to_bin_size(self.bin_edges)  # type: ignore


# todo: add option for the following bin specifications.
# Bin specifications from an appropriate class or as defined in :func:`numpy.histogramdd`:
# The number of bins for all dimensions n.
# The number of bins for each dimension (nx, ny)
# A sequence of arrays ((edge_x1, edge_x2), (edge_y1, edge_y2)) describing the monotonically
# increasing bin edges along each dimension.
class Bins:
    """
    Bin definitions to be used in histogram and render functions.
    Bin edges are continuous, contiguous and monotonic.
    Bins can be instantiated from specifications for `bins` or `bin_edges`
    or for one of `n_bins` or `bin_size` in combination with `bin_range`.
    One and only one of (`bins`, `bin_edges`, `n_bins`, `bin_size`) must be different
    from None in any instantiating function.
    To pass bin specifications to other functions use an instance of
    `Bins` or `bin_edges`.

    Parameters
    ----------
    bins : Bins | boost_histogram.axis.Axis | boost_histogram.axis.AxesTuple | None
        Specific class specifying the bins.
    bin_edges : Sequence[float] | Sequence[Sequence[float]] | None
        Bin edges for all or each dimension
        with shape (dimension, n_bin_edges).
    bin_range : tuple[float, float] | Sequence[float] | Sequence[Sequence[float]]
        Minimum and maximum edge for all or each dimensions
        with shape (2,) or (dimension, 2).
    n_bins : int | Sequence[int] | None
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size : float | Sequence[float] | Sequence[Sequence[float]] | None
        The size of bins for all or each bin and for all or each dimension
        with shape (dimension,) or (dimension, n_bins).
        5 would describe bin_size of 5 for all bins in all dimensions.
        ((2, 5),) yield bins of size (2, 5) for one dimension.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other
        dimension.
        ((2, 5), (1, 3)) yields bins of size (2, 5) for one dimension and
        (1, 3) for the other dimension.
        To specify arbitrary sequence of `bin_size` use `bin_edges` instead.
    labels : list[str] | None
        Names for each bin axis with shape (dimension,)
    extend_range : bool | None
        If for equally-sized bins the final bin_edge is different from the
        maximum bin_range, the last bin_edge will be smaller than the maximum
        bin_range but all bins are equally-sized (None);
        the last bin_edge will be equal to the maximum bin_range but bins
        are not equally-sized (False);
        the last bin_edge will be larger than the maximum bin_range but all
        bins are equally-sized (True).

        If for variable-sized bins the final bin_edge is different from the
        maximum bin_range, the last bin_edge will be smaller than the maximum
        bin_range (None);
        the last bin_edge will be equal to the maximum bin_range (False);
        the last bin_edge will be larger than the maximum bin_range but taken
        from the input sequence (True).

    Attributes
    ----------
    dimension : int
        The number of dimensions for which bins are provided.
    bin_range : tuple[tuple[float, float], ...]
        Minimum and maximum edge for each dimension with shape (dimension, 2).
    bin_edges : tuple[npt.NDArray[np.float64], ...]
        Array(s) with bin edges for each dimension with shape (dimension,)
    n_bins : tuple[int, ...]
        Number of bins for each dimension with shape (dimension,)
    bin_size : tuple[float, ...] | tuple[npt.NDArray[np.float64], ...]
        Size of bins for each dimension with shape (dimension,)
        or with shape (dimension, n_bins).
    bin_centers :  tuple[npt.NDArray[np.float64], ...]
        Array(s) with bin centers for all or each dimension
        with shape (dimension,).
    labels : list[str] | None
        Names for each bin axis.
    boost_histogram_axes : boost_histogram.axis.AxesTuple
        Axis definitions for boost-histogram
    """

    def __init__(
        self,
        bins: Bins | bh.axis.Axis | bh.axis.AxesTuple | None = None,
        n_bins: int | Sequence[int] | None = None,
        bin_size: float | Sequence[float] | Sequence[Sequence[float]] | None = None,
        bin_edges: Sequence[float] | Sequence[Sequence[float]] | None = None,
        bin_range: (
            tuple[float, float] | Sequence[float] | Sequence[Sequence[float]] | None
        ) = None,
        labels: list[str] | None = None,
        extend_range: bool | None = None,
    ) -> None:
        self._bins: (
            Bins
            | _BinsFromBoostHistogramAxis
            | _BinsFromNumber
            | _BinsFromSize
            | _BinsFromEdges
        )
        self._labels: list[str] | None

        # check for correct inputs
        excluding_parameter = (bins, n_bins, bin_size, bin_edges)
        excluding_parameter_strings = ("bins", "n_bins", "bin_size", "bin_edges")
        n_inputs = sum(param is not None for param in excluding_parameter)
        if n_inputs != 1:
            raise ValueError(
                f"One and only one of {excluding_parameter_strings} "
                f"must be different from None."
            )

        # inject builder class
        if bins is not None:
            if bin_range is not None:
                raise ValueError(
                    "The parameter `bin_range` is derived from bins class "
                    "and must be None."
                )
            if isinstance(bins, Bins):
                self._bins = bins
            if isinstance(bins, (bh.axis.Axis, bh.axis.AxesTuple)):
                self._bins = _BinsFromBoostHistogramAxis(bins)
        elif n_bins is not None:
            self._bins = _BinsFromNumber(n_bins, bin_range)  # type: ignore
        elif bin_size is not None:
            self._bins = _BinsFromSize(bin_size, bin_range, extend_range)  # type: ignore
        elif bin_edges is not None:
            if bin_range is not None:
                raise ValueError(
                    "The parameter `bin_range` is derived from `bin_edges` "
                    "and must be None."
                )
            self._bins = _BinsFromEdges(bin_edges)

        self._bin_centers: tuple[npt.NDArray[np.float64], ...] | None = None
        self.labels = labels
        self._boost_histogram_axes: bh.axis.AxesTuple | None = None

    @property
    def dimension(self) -> int:
        return self._bins.dimension

    @property
    def bin_edges(self) -> tuple[npt.NDArray[np.float64], ...]:
        return_value: tuple[npt.NDArray[np.float64], ...] = self._bins.bin_edges
        return return_value

    @property
    def n_bins(self) -> tuple[int, ...]:
        return self._bins.n_bins

    @property
    def bin_size(self) -> tuple[float, ...] | tuple[npt.NDArray[np.float64], ...]:
        return self._bins.bin_size

    @property
    def bin_range(self) -> tuple[tuple[float, float], ...]:
        return self._bins.bin_range

    @property
    def bin_centers(self) -> tuple[npt.NDArray[np.float64], ...]:
        if self._bin_centers is None:
            self._bin_centers = getattr(self._bins, "bin_centers", None)
            if self._bin_centers is None:
                self._bin_centers = _bin_edges_to_bin_centers(self.bin_edges)  # type: ignore
        return_value: tuple[npt.NDArray[np.float64], ...] = self._bin_centers
        return return_value

    @property
    def labels(self) -> list[str] | None:
        if self._labels is None:
            self._labels = getattr(self._bins, "labels", None)
        return self._labels

    @labels.setter
    def labels(self, value: str | Sequence[str] | None) -> None:
        if value is None:
            self._labels = None
        elif isinstance(value, str):
            self._labels = [value]
        elif isinstance(value, (tuple, list)):
            self._labels = list(value)
        else:
            raise TypeError("`labels` must be str or list of str or None.")

        if self._labels is not None and len(self.labels) != self.dimension:  # type: ignore
            self._labels = None
            raise ValueError("`labels` must have a length of `dimension`.")

    def __getattr__(self, attr: str) -> Any:
        if attr.startswith("__") and attr.endswith(
            "__"
        ):  # this is needed to enable pickling
            raise AttributeError
        return getattr(self._bins, attr)

    @property
    def is_equally_sized(self) -> tuple[bool, ...]:
        """True for each dimension if all bins are of the same size."""
        if _is_1d_array_of_scalar(self.bin_size):
            return tuple(True for _ in self.bin_size)
        elif _is_2d_array_of_1d_array_of_scalar(self.bin_size):
            return tuple(all(np.isclose(bs, bs[0])) for bs in self.bin_size)  # type: ignore

        elif _is_2d_inhomogeneous_array(self.bin_size):
            result = []
            for bs in self.bin_size:
                if _is_scalar(bs):
                    result.append(True)
                else:
                    result.append(all(np.isclose(bs, bs[0])))  # type: ignore
            return tuple(result)
        else:
            raise TypeError

    def equalize_bin_size(self) -> Bins:
        """
        Return a new instance of `Bins` with bin_size set equal to the
        first bin_size element in each dimension
        and extend_range=None.
        """
        new_bin_size = []
        for bs in self.bin_size:
            try:
                new_bin_size.append(bs[0])  # type: ignore
            except IndexError:
                new_bin_size.append(bs)
        return Bins(bin_size=new_bin_size, bin_range=self.bin_range, extend_range=None)

    @property
    def boost_histogram_axes(self) -> bh.axis.AxesTuple:
        """Axis definitions for boost-histogram"""
        if self._boost_histogram_axes is None:
            axes = []
            for index in range(self.dimension):
                if self.is_equally_sized[index]:
                    axis = bh.axis.Regular(self.n_bins[index], *self.bin_range[index])
                else:
                    axis = bh.axis.Variable(self.bin_edges[index])  # type: ignore
                axes.append(axis)
            self._boost_histogram_axes = bh.axis.AxesTuple(axes)
        return self._boost_histogram_axes


def _histogram_fast_histogram(data: npt.ArrayLike, bins: Bins) -> npt.NDArray[np.int64]:
    """
    Provide histogram with counts in each bin.

    Parameters
    ----------
    data
        Coordinate values with shape (dimensions, n_points) to be binned
    bins
        The bin specification

    Returns
    -------
    npt.NDArray[np.int64]
    """
    data = np.asarray(data)
    if data.shape[0] == 1:
        img: npt.NDArray[np.int64] = fast_histogram.histogram1d(
            data, range=bins.bin_range[0], bins=bins.n_bins[0]
        )
    elif data.shape[0] == 2:
        img = fast_histogram.histogram2d(*data, range=bins.bin_range, bins=bins.n_bins)
    else:
        raise TypeError("Dimension of data must be 1 or 2.")
    return img


def _histogram_boost_histogram(
    data: npt.ArrayLike, bins: Bins
) -> npt.NDArray[np.int64]:
    """
    Provide histogram with counts in each bin.

    Parameters
    ----------
    data
        Coordinate values with shape (n_dimensions, n_points) to be binned
    bins
        The bin specification

    Returns
    -------
    npt.NDArray[np.int64]
    """
    hist = bh.Histogram(*bins.boost_histogram_axes).fill(*data)  # type: ignore
    img: npt.NDArray[np.int64] = hist.view()
    return img


def _histogram_mean_fast_histogram(
    data: npt.ArrayLike, bins: Bins, values: npt.ArrayLike
) -> npt.NDArray[np.float64]:
    """
    Provide histogram with averaged values for all counts in each bin.

    Parameters
    ----------
    data
        Coordinate values with shape (n_dimensions, n_points) to be binned
    bins
        The bin specification
    values
        Values with shape (n_points,) to be averaged in each bin

    Returns
    -------
    npt.NDArray[np.float64]
    """
    data = np.asarray(data)
    if data.shape[0] == 1:
        hist: npt.NDArray[np.float64] = fast_histogram.histogram1d(
            data, range=bins.bin_range[0], bins=bins.n_bins[0]
        )
        hist_w = fast_histogram.histogram1d(
            data, range=bins.bin_range[0], bins=bins.n_bins[0], weights=values
        )
    elif data.shape[0] == 2:
        hist = fast_histogram.histogram2d(*data, range=bins.bin_range, bins=bins.n_bins)
        hist_w = fast_histogram.histogram2d(
            *data, range=bins.bin_range, bins=bins.n_bins, weights=values
        )
    else:
        raise TypeError("Dimension of data must be 1 or 2.")

    with np.errstate(divide="ignore", invalid="ignore"):
        hist = np.true_divide(hist_w, hist)
        hist[hist == np.inf] = 0
        # hist = np.nan_to_num(hist)
    return hist


def _histogram_mean_boost_histogram(
    data: npt.ArrayLike, bins: Bins, values: npt.ArrayLike
) -> npt.NDArray[np.float64]:
    """
    Provide histogram with averaged values for all counts in each bin.

    Parameters
    ----------
    data : npt.ArrayLike
        Coordinate values with shape (n_dimensions, n_points) to be binned
    bins
        The bin specification
    values
        Values with shape (n_points,) to be averaged in each bin

    Returns
    -------
    npt.NDArray[np.float64]
    """
    hist = bh.Histogram(*bins.boost_histogram_axes, storage=bh.storage.Mean()).fill(  # type: ignore
        *data, sample=values
    )
    # bh.Histogram yields zero for mean values in bins with zero counts
    mean_values: npt.NDArray[np.float64] = hist.values()
    mask = hist.counts() == 0
    mean_values[mask] = np.nan
    return mean_values


def histogram(
    locdata: LocData,
    loc_properties: str | Iterable[str] | None = None,
    other_property: str | None = None,
    bins: Bins | bh.axis.Axis | bh.axis.AxesTuple | None = None,
    n_bins: int | Sequence[int] | None = None,
    bin_size: float | Sequence[float] | Sequence[Sequence[float]] | None = None,
    bin_edges: Sequence[float] | Sequence[Sequence[float]] | None = None,
    bin_range: (
        tuple[float, float]
        | Sequence[float]
        | Sequence[Sequence[float]]
        | Literal["zero", "link"]
        | None
    ) = None,
) -> tuple[npt.NDArray[np.int64 | np.float64], Bins, list[str]]:
    """
    Make histogram of loc_properties (columns in `locdata.data`)
    by binning all localizations
    or averaging other_property within each bin.

    Parameters
    ----------
    locdata
        Localization data.
    loc_properties
        Localization properties to be grouped into bins.
        If None The coordinate_values of locdata are used.
    other_property
        Localization property that is averaged in each pixel.
        If None localization counts are shown.
    bins
        The bin specification as defined in :class:`Bins`
    bin_edges
        Bin edges for all or each dimension
        with shape (dimension, n_bin_edges).
    bin_range
        Minimum and maximum edge for all or each dimensions
        with shape (2,) or (dimension, 2).
        If None (min, max) ranges are determined from data and returned;
        if 'zero' (0, max) ranges with max determined from data are returned.
        if 'link' (min_all, max_all) ranges with min and max determined from
        all combined data are returned.
    n_bins
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size
        The size of bins for all or each bin and for all or each dimension
        with shape (dimension,) or (dimension, n_bins).
        5 would describe bin_size of 5 for all bins in all dimensions.
        ((2, 5),) yield bins of size (2, 5) for one dimension.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other
        dimension.
        ((2, 5), (1, 3)) yields bins of size (2, 5) for one dimension and
        (1, 3) for the other dimension.
        To specify arbitrary sequence of `bin_size` use `bin_edges` instead.

    Returns
    -------
    namedtuple('Histogram', "data bins labels"): (npt.NDArray[np.int64 | np.float64], Bins, list[str])
    """
    labels_ = _check_loc_properties(locdata, loc_properties)
    data = locdata.data[labels_].values.T
    img: npt.NDArray[np.int64 | np.float64]

    if (
        (bin_range is None or isinstance(bin_range, str))
        and bin_edges is None
        and bins is None
    ):
        bin_range_ = ranges(locdata, loc_properties=labels_, special=bin_range)
    else:
        bin_range_ = bin_range  # type: ignore

    try:
        bins = Bins(
            bins=bins,
            n_bins=n_bins,
            bin_size=bin_size,
            bin_edges=bin_edges,
            bin_range=bin_range_,  # type: ignore
            labels=labels_,
        )
    except ValueError as exc:  # the error is raised again only to adapt the message.
        raise ValueError(
            "Bin dimension and len of `loc_properties` is incompatible."
        ) from exc

    if other_property is None:
        # histogram data by counting points
        if data.shape[0] == 2:
            # we are using fast-histogram for 2D since it is even faster
            # than boost_histogram
            img = _histogram_fast_histogram(data, bins)
        elif data.shape[0] == 1 or data.shape[0] == 3:
            img = _histogram_boost_histogram(data, bins)
        else:
            raise TypeError(
                "loc_properties must contain a string or a list with 1, 2 or 3 "
                "elements."
            )
        labels_.append("counts")

    elif other_property in locdata.data.columns:
        # histogram data by averaging values
        values = locdata.data[other_property].values
        if data.shape[0] == 2:
            img = _histogram_mean_fast_histogram(data=data, bins=bins, values=values)  # type: ignore
        elif data.shape[0] == 1 or data.shape[0] == 3:
            img = _histogram_mean_boost_histogram(data=data, bins=bins, values=values)  # type: ignore
        else:
            raise TypeError("No more than 3 elements in loc_properties are allowed.")
        labels_.append(other_property)
    else:
        raise TypeError(
            f"Parameter for `other_property` {other_property} is not a valid property "
            f"name."
        )

    Histogram = namedtuple("Histogram", "data bins labels")
    return Histogram(img, bins, labels_)


def _accumulate_1d(
    data: npt.ArrayLike,
    bin_edges: npt.ArrayLike,
    return_data: bool = False,
    return_counts: bool = False,
) -> tuple[
    npt.NDArray[np.int64], list[int], list[int] | None, npt.NDArray[np.int64] | None
]:
    """
    Bin data and collect data elements contained in each bin.
    The returned `bin_indices` refer to the given bins including index[0] for
    underflow data and index[n_bins] for overflow data.

    Parameters
    ----------
    data
        Data array of shape (n_points,) or (n_points, dimensions)
        All points are binned with regard to the first dimension.
    bin_edges
        Array of bin edges for corresponding dimension.
    return_data
        If true, grouped data elements are returned.
    return_counts
        If true, counts (number of elements per bin) are returned.

    Note
    ----
    Even though the returned data groups are sorted according to the bins,
    the data within groups is not sorted.

    Returns
    -------
    tuple[npt.NDArray[np.int64], list[int], list[int] | None, npt.NDArray[np.int64] | None]
        bin_indices, data_indices, collection, counts.
    """
    data_ = np.array(data)
    if data_.ndim > 1:
        data_ = data_[:, 0]
    # identify bins indices
    bin_identifier = np.digitize(data_, bins=bin_edges)  # type: ignore
    # bin_identifier 0 and n_bins represent out of bounds data

    # sort data
    sorted_indices = np.argsort(bin_identifier, kind="stable")

    # group data
    bin_indices, n_elements = np.unique(bin_identifier, return_counts=True)
    # bin_indices (like bin_identifier) 0 and n_bins represent out of bounds data

    cumsum = np.cumsum(n_elements)
    start_indices = np.insert(cumsum[:-1], 0, 0)
    stop_indices = cumsum
    data_indices = [
        sorted_indices[start:stop] for start, stop in zip(start_indices, stop_indices)
    ]

    collection = [data[indices_] for indices_ in data_indices] if return_data else None  # type: ignore
    counts = n_elements if return_counts else None

    return bin_indices, data_indices, collection, counts  # type: ignore


def _accumulate_2d(
    data: npt.ArrayLike,
    bin_edges: tuple[npt.ArrayLike, ...],
    return_data: bool = False,
    return_counts: bool = False,
) -> tuple[
    npt.NDArray[np.int64], list[int], list[int] | None, npt.NDArray[np.int64] | None
]:
    """
    Bin data and collect data elements contained in each bin.
    All points are binned with regard to the first and second dimension.
    The returned `bin_indices` refer to the given bins
    including index[0] for underflow data
    and index[n_bins] for overflow data.

    Parameters
    ----------
    data
        Data array of shape (n_points, dimensions)
    bin_edges
        Array of bin edges for corresponding dimensions.
    return_data
        If true, grouped data elements are returned.
    return_counts
        If true, counts (number of elements per bin) are returned.

    Returns
    -------
    tuple[npt.NDArray[np.int64], list[int], list[int] | None, npt.NDArray[np.int64] | None]
        bin_indices, data_indices, collection, counts.
    """
    data_ = np.array(data)
    if data_.size == 0:
        bin_indices = np.array([])
        data_indices: list[int] = []
        collection: list[int] | None | None = [] if return_data else None
        counts = np.array([]) if return_counts else None
        return bin_indices, data_indices, collection, counts

    # accumulate first dimension
    bin_indices_first_dim, data_indices_first_dim, _, _ = _accumulate_1d(
        data=data_[:, 0], bin_edges=bin_edges[0]
    )

    # traverse groups
    bin_indices_ = []
    data_indices = []
    counts_ = []
    for bin_index_first_dim_, data_indices_first_dim_ in zip(
        bin_indices_first_dim, data_indices_first_dim
    ):
        grouped_data_ = data_[:, 1][data_indices_first_dim_]
        bin_indices_group, data_indices_group, _, counts_group = _accumulate_1d(
            data=grouped_data_, bin_edges=bin_edges[1], return_counts=return_counts
        )

        # form multi-dimensional bin_indices
        first = np.repeat(bin_index_first_dim_, len(bin_indices_group))
        new_bin_indices = np.vstack([first, bin_indices_group]).T
        bin_indices_.append(new_bin_indices)
        counts_.append(counts_group)

        # form multi-dimensional data_indices
        new_data_indices = [
            data_indices_first_dim_[idxs] for idxs in data_indices_group  # type: ignore
        ]
        data_indices.extend(new_data_indices)

    bin_indices = np.concatenate(bin_indices_)
    counts = np.concatenate(counts_) if return_counts else None  # type: ignore
    collection = [data_[idxs] for idxs in data_indices] if return_data else None

    return bin_indices, data_indices, collection, counts



"""

A class to carry localization data.

"""

from __future__ import annotations

import copy
import logging
import os
import sys
import warnings
from collections.abc import Callable, Iterable, Sequence
from itertools import accumulate
from typing import TYPE_CHECKING, Any, BinaryIO, Literal, TypeVar

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

import numpy as np
import numpy.typing as npt
import pandas as pd
from google.protobuf import json_format

try:
    from scipy.spatial import QhullError
except ImportError:
    from scipy.spatial.qhull import QhullError  # needed for Python 3.7

import locan.data.hulls
from locan import (  # is required to use locdata_id as global variable  # noqa: F401
    locdata_id,
)
from locan.constants import PROPERTY_KEYS, PropertyKey
from locan.data import metadata_pb2
from locan.data.locdata_utils import _dataframe_to_pandas, _get_linked_coordinates
from locan.data.metadata_utils import (
    _modify_meta,
    merge_metadata,
    metadata_to_formatted_string,
)
from locan.data.regions.region import Region, RoiRegion

if TYPE_CHECKING:
    from locan.locan_types import DataFrame  # noqa F401

__all__: list[str] = ["LocData"]

logger = logging.getLogger(__name__)

T_LocData = TypeVar("T_LocData", bound="LocData")


class LocData:
    """
    This class carries localization data, aggregated properties and meta data.

    Data consist of individual elements being either localizations or other
    `LocData` objects.
    Both, localizations and `Locdata` objects have properties.
    Properties come from the original data or are added by analysis procedures.

    Parameters
    ----------
    references
        A `LocData` reference or an array with references to `LocData` objects
        referring to the selected localizations in dataset.
    dataframe
        Dataframe with localization data.
    indices
        Indices for dataframe in references that makes up the data.
        `indices` refers to index label, not position.
    meta
        Metadata about the current dataset and its history.

    Attributes
    ----------
    references : LocData | list[LocData] | None
        A LocData reference or an array with references to LocData objects
        referring to the selected localizations in dataframe.
    dataframe : pandas.DataFrame
        Dataframe with localization data.
    indices : slice | list[int] | None
        Indices for dataframe in references that makes up the data.
    meta : locan.data.metadata_pb2.Metadata
        Metadata about the current dataset and its history.
    properties : dict[str, Any]
        List of properties generated from data.
    coordinate_keys : list[str]
        The available coordinate properties.
    uncertainty_keys : list[str]
        The available uncertainty properties.
    dimension : int
        Number of coordinates available for each localization
        (i.e. size of `coordinate_keys`).
    """

    count = 0
    """int: A counter for counting LocData instantiations (class attribute)."""

    def __init__(
        self,
        references: LocData | Iterable[LocData] | None = None,
        dataframe: pd.DataFrame | None = None,
        indices: (
            int
            | list[int | bool]
            | npt.NDArray[np.int64 | np.bool_]
            | slice
            | pd.Index[int]
            | None
        ) = None,
        meta: (
            metadata_pb2.Metadata
            | dict[str, Any]
            | str
            | bytes
            | os.PathLike[Any]
            | BinaryIO
            | None
        ) = None,
    ):
        self.__class__.count += 1

        self.references: LocData | Iterable[LocData] | None = references
        self.dataframe: pd.DataFrame = (
            pd.DataFrame() if dataframe is None else dataframe
        )
        self.indices: (
            int
            | list[int | bool]
            | npt.NDArray[np.int64 | np.bool_]
            | slice
            | pd.Index[int]
            | None
        ) = indices
        self.meta: metadata_pb2.Metadata = metadata_pb2.Metadata()
        self.properties: dict[str, Any] = {}

        # regions and hulls
        self._region: Region | None = None
        self._bounding_box: locan.data.hulls.BoundingBox | None = None
        self._oriented_bounding_box: locan.data.hulls.OrientedBoundingBox | None = None
        self._convex_hull: locan.data.hulls.ConvexHull | None = None
        self._alpha_shape: locan.data.hulls.AlphaShape | None = None
        self._inertia_moments: locan.process.properties.misc.InertiaMoments | None = (
            None
        )

        self.dimension: int = len(self.coordinate_keys)

        self._update_properties()

        # meta
        global locdata_id
        locdata_id += 1
        self.meta.identifier = str(locdata_id)

        self.meta.creation_time.GetCurrentTime()
        self.meta.source = metadata_pb2.DESIGN
        self.meta.state = metadata_pb2.RAW

        self.meta.element_count = len(self.data.index)
        if "frame" in self.data.columns:
            self.meta.frame_count = len(self.data["frame"].unique())

        self.meta = merge_metadata(metadata=self.meta, other_metadata=meta)

    @property
    def coordinate_keys(self) -> list[str]:
        return [
            label_
            for label_ in PropertyKey.coordinate_keys()
            if label_ in self.data.columns
        ]

    @property
    def uncertainty_keys(self) -> list[str]:
        return [
            label_
            for label_ in PropertyKey.uncertainty_keys()
            if label_ in self.data.columns
        ]

    def _update_properties(
        self, update_function: dict[str, Callable[..., Any]] | None = None
    ) -> Self:
        """
        Compute properties from localization data.

        For each loc_property in update_function the supplied callable is used.
        If None the following functions are used if loc_properties are available:

        coordinates and corresponding uncertainties: weighted_mean_variance
        `intensity`: sum
        `local_background`: mean
        `frame`: min

        Parameters
        ----------
        update_function
            mapping of localization property onto callable to compute
            property from corresponding localization data

        Returns
        -------
        Self
        """
        self.properties = dict()
        self.properties["localization_count"] = len(self.data.index)

        properties_for_update = [
            loc_property_
            for loc_property_ in [
                *self.coordinate_keys,
                "frame",
                "intensity",
                "local_background",
            ]
            if loc_property_ in self.data.columns
        ]

        if update_function is not None:
            properties_for_update = [
                loc_property_
                for loc_property_ in properties_for_update
                if loc_property_ not in update_function.keys()
            ]

        # localization coordinates
        if all(c_label_ in properties_for_update for c_label_ in self.coordinate_keys):
            self.properties.update(_get_linked_coordinates(locdata=self.data))

        if "intensity" in properties_for_update:
            self.properties["intensity"] = np.sum(self.data["intensity"])

        if "local_background" in properties_for_update:
            self.properties["local_background"] = np.mean(self.data["local_background"])

        if "frame" in properties_for_update:
            self.properties["frame"] = np.min(self.data["frame"])

        if update_function is not None:
            for loc_property_, function_ in update_function.items():
                self.properties[loc_property_] = function_(self.data[loc_property_])

        self.bounding_box  # update self._bounding_box  # noqa B018
        return self

    def __del__(self) -> None:
        """Updating the counter upon deletion of class instance."""
        self.__class__.count -= 1

    def __len__(self) -> int:
        """
        Return the length of data, i.e. the number of elements
        (localizations or collection elements).
        """
        return len(self.data.index)

    def __getstate__(self) -> dict[str, Any]:
        """Modify pickling behavior."""
        # Copy the object's state from self.__dict__ to avoid modifying the original state.
        state = self.__dict__.copy()
        # Serialize the unpicklable protobuf entries.
        json_string = json_format.MessageToJson(
            self.meta, always_print_fields_with_no_presence=False
        )
        state["meta"] = json_string
        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        """Modify pickling behavior."""
        # Restore instance attributes.
        self.__dict__.update(state)
        # Restore protobuf class for meta attribute
        self.meta = metadata_pb2.Metadata()
        self.meta = json_format.Parse(state["meta"], self.meta)

    def __copy__(self) -> LocData:
        """
        Create a shallow copy of locdata (keeping all references) with the
        following exceptions:
        (i) The class variable `count` is increased for the copied LocData
        object.
        (ii) Metadata keeps the original meta.creation_time while
        meta.modification_time and meta.history is updated.
        """
        new_locdata = LocData(self.references, self.dataframe, self.indices, meta=None)
        new_locdata._region = self._region
        # meta
        meta_ = _modify_meta(
            self, new_locdata, function_name="LocData.copy", parameter=None, meta=None
        )
        new_locdata.meta = meta_
        return new_locdata

    def __deepcopy__(self, memodict: dict[Any, Any] | None = None) -> LocData:
        """
        Create a deep copy of locdata (including all references) with the
        following exceptions:
        (i) The class variable `count` is increased for all deepcopied LocData
        objects.
        (ii) Metadata keeps the original meta.creation_time while
        meta.modification_time and meta.history is updated.
        """
        if memodict is None:
            memodict = {}
        new_locdata = LocData(
            copy.deepcopy(self.references, memodict),
            copy.deepcopy(self.dataframe, memodict),
            copy.deepcopy(self.indices, memodict),
            meta=None,
        )
        new_locdata._region = self._region
        # meta
        meta_ = _modify_meta(
            self,
            new_locdata,
            function_name="LocData.deepcopy",
            parameter=None,
            meta=None,
        )
        new_locdata.meta = meta_
        return new_locdata

    @property
    def bounding_box(self) -> locan.data.hulls.BoundingBox | None:
        """
        Hull object: Return an object representing the axis-aligned
        minimal bounding box.
        """
        if self._bounding_box is None:
            try:
                self._bounding_box = locan.data.hulls.BoundingBox(self.coordinates)
            except ValueError:
                warnings.warn(
                    "Properties related to bounding box could not be computed.",
                    UserWarning,
                    stacklevel=1,
                )
        self._update_properties_bounding_box()
        return self._bounding_box

    def _update_properties_bounding_box(self) -> None:
        if self._bounding_box is not None:
            self.properties["region_measure_bb"] = self._bounding_box.region_measure
            if self._bounding_box.region_measure:
                self.properties["localization_density_bb"] = (
                    self.properties["localization_count"]
                    / self._bounding_box.region_measure
                )
            if self._bounding_box.subregion_measure:
                self.properties["subregion_measure_bb"] = (
                    self._bounding_box.subregion_measure
                )

    @property
    def convex_hull(self) -> locan.data.hulls.ConvexHull | None:
        """
        Hull object: Return an object representing the convex hull of all
        localizations.
        """
        if self._convex_hull is None:
            try:
                self._convex_hull = locan.data.hulls.ConvexHull(self.coordinates)
            except (TypeError, QhullError):
                warnings.warn(
                    "Properties related to convex hull could not be computed.",
                    UserWarning,
                    stacklevel=1,
                )
        self._update_properties_convex_hull()
        return self._convex_hull

    def _update_properties_convex_hull(self) -> None:
        if self._convex_hull is not None:
            self.properties["region_measure_ch"] = self._convex_hull.region_measure
            if self._convex_hull.region_measure:
                self.properties["localization_density_ch"] = (
                    self.properties["localization_count"]
                    / self._convex_hull.region_measure
                )
            if self._convex_hull.subregion_measure:
                self.properties["subregion_measure_ch"] = (
                    self._convex_hull.subregion_measure
                )

    @property
    def oriented_bounding_box(self) -> locan.data.hulls.OrientedBoundingBox | None:
        """
        Hull object: Return an object representing the oriented minimal
        bounding box.
        """
        if self._oriented_bounding_box is None:
            try:
                self._oriented_bounding_box = locan.data.hulls.OrientedBoundingBox(
                    self.coordinates
                )
            except TypeError:
                warnings.warn(
                    "Properties related to oriented bounding box could not be computed.",
                    UserWarning,
                    stacklevel=1,
                )
        self._update_properties_oriented_bounding_box()
        return self._oriented_bounding_box

    def _update_properties_oriented_bounding_box(self) -> None:
        if self._oriented_bounding_box is not None:
            self.properties["region_measure_obb"] = (
                self._oriented_bounding_box.region_measure
            )
            if self._oriented_bounding_box.region_measure:
                self.properties["localization_density_obb"] = (
                    self.properties["localization_count"]
                    / self._oriented_bounding_box.region_measure
                )
            self.properties["orientation_obb"] = self._oriented_bounding_box.angle
            self.properties["circularity_obb"] = self._oriented_bounding_box.elongation

    @property
    def alpha_shape(self) -> locan.data.hulls.AlphaShape | None:
        """
        Hull object: Return an object representing the alpha-shape of all
        localizations.
        """
        return self._alpha_shape

    def update_alpha_shape(self, alpha: float) -> Self:
        """
        Compute the alpha shape for specific `alpha` and update
        `self.alpha_shape`.

        Parameters
        ----------
        alpha
            Alpha parameter specifying a unique alpha complex.

        Returns
        -------
        Self
            The modified object
        """
        try:
            if self._alpha_shape is None:
                self._alpha_shape = locan.data.hulls.AlphaShape(
                    points=self.coordinates, alpha=alpha
                )
            else:
                self._alpha_shape.alpha = alpha
        except TypeError:
            warnings.warn(
                "Properties related to alpha shape could not be computed.",
                UserWarning,
                stacklevel=1,
            )
        self._update_properties_alpha_shape()
        return self

    def _update_properties_alpha_shape(self) -> None:
        if self._alpha_shape is not None:
            self.properties["region_measure_as"] = self._alpha_shape.region_measure
            try:
                self.properties["localization_density_as"] = (
                    self._alpha_shape.n_points_alpha_shape
                    / self._alpha_shape.region_measure
                )
            except ZeroDivisionError:
                self.properties["localization_density_as"] = float("nan")

    def update_alpha_shape_in_references(self, alpha: float) -> Self:
        """
        Compute the alpha shape for each element in `locdata.references` and
        update `locdata.dataframe`.

        Parameters
        ----------
        alpha
            Alpha parameter specifying a unique alpha complex.

        Returns
        -------
        Self
            The modified object
        """
        if isinstance(self.references, list):
            for reference in self.references:
                reference.update_alpha_shape(alpha=alpha)
            new_df = pd.DataFrame(
                [reference.properties for reference in self.references]
            )
            new_df.index = self.data.index
            if self.dataframe is None:
                self.dataframe = new_df
            else:
                self.dataframe.update(new_df)
            new_columns = [
                column for column in new_df.columns if column in self.dataframe.columns
            ]
            new_df.drop(columns=new_columns, inplace=True, errors="ignore")
            self.dataframe = pd.concat([self.dataframe, new_df], axis=1)
        return self

    @property
    def inertia_moments(self) -> locan.process.properties.misc.InertiaMoments | None:
        """
        Inertia moments are returned as computed by
        :func:`locan.process.properties.inertia_moments`.
        """
        if self._inertia_moments is None:
            try:
                self._inertia_moments = locan.process.properties.inertia_moments(
                    self.coordinates
                )
            except TypeError:
                warnings.warn(
                    "Properties related to inertia_moments could not be computed.",
                    UserWarning,
                    stacklevel=1,
                )
        self._update_properties_inertia_moments()
        return self._inertia_moments

    def _update_properties_inertia_moments(self) -> None:
        if self._inertia_moments is not None:
            self.properties["orientation_im"] = self._inertia_moments.orientation
            self.properties["circularity_im"] = self._inertia_moments.eccentricity

    def update_inertia_moments_in_references(self) -> Self:
        """
        Compute inertia_moments for each element in locdata.references and
        update locdata.dataframe.

        Returns
        -------
        Self
            The modified object
        """
        if isinstance(self.references, list):
            for reference in self.references:
                reference.inertia_moments  # request property to update  # noqa B018
            new_df = pd.DataFrame(
                [reference.properties for reference in self.references]
            )
            new_df.index = self.data.index
            if self.dataframe is None:
                self.dataframe = new_df
            else:
                self.dataframe.update(new_df)
            new_columns = [
                column for column in new_df.columns if column in self.dataframe.columns
            ]
            new_df.drop(columns=new_columns, inplace=True, errors="ignore")
            self.dataframe = pd.concat([self.dataframe, new_df], axis=1)
        return self

    @property
    def region(self) -> Region | None:
        """RoiRegion object: Return the region that supports all localizations."""
        return self._region

    @region.setter
    def region(self, region: Region | None) -> None:
        if region is not None:
            if region.dimension != self.dimension:
                raise TypeError(
                    "Region dimension and coordinates dimension must be identical."
                )
            elif len(self) != len(region.contains(self.coordinates)):
                logger.warning("Not all coordinates are within region.")

        if isinstance(region, (Region, RoiRegion)) or region is None:
            self._region = region

        elif isinstance(
            region, dict
        ):  # legacy code to deal with deprecated RoiLegacy_0
            region_ = RoiRegion(**region)
            if region_ is not None:
                if region_.dimension != self.dimension:
                    raise TypeError(
                        "Region dimension and coordinates dimension must be identical."
                    )
                elif len(self) != len(region_.contains(self.coordinates)):
                    logger.warning("Not all coordinates are within region.")
            self._region = region_

        else:
            raise TypeError

        # property for region measures
        if self._region is not None:
            if self._region.region_measure:
                self.properties["region_measure"] = self._region.region_measure
                self.properties["localization_density"] = (
                    self.meta.element_count / self._region.region_measure
                )
            if self._region.subregion_measure:
                self.properties["subregion_measure"] = self._region.subregion_measure

    @property
    def data(self) -> pd.DataFrame:
        """
        pandas.DataFrame: Return all elements either copied from the reference
        or referencing the current dataframe.
        """
        if isinstance(self.references, LocData):
            # we refer to the localization data by its index label, not position
            # in other words we decided not to use iloc but loc
            # df = self.references.data.loc[self.indices]  ... but this does not work in pandas.
            # also see:
            # https://pandas.pydata.org/pandas-docs/stable/user_guide/indexing.html#deprecate-loc-reindex-listlike
            try:
                df = self.references.data.loc[self.indices]  # type: ignore
            except KeyError:
                df = self.references.data.loc[
                    self.references.data.index.intersection(self.indices)  # type: ignore
                ]
            df = pd.merge(
                df, self.dataframe, left_index=True, right_index=True, how="outer"
            )
            return df
        else:
            return self.dataframe

    @property
    def coordinates(self) -> npt.NDArray[np.float64]:
        """npt.NDArray[float]: Return all coordinate values."""
        return_value: npt.NDArray[np.float64] = self.data[self.coordinate_keys].values
        return return_value

    @property
    def centroid(self) -> npt.NDArray[np.float64]:
        """
        npt.NDArray[np.float64]: Return coordinate values of the centroid
        (being the property values for all coordinate labels).
        """
        return np.array(
            [
                self.properties[coordinate_label]
                for coordinate_label in self.coordinate_keys
            ]
        )

    @classmethod
    def from_dataframe(
        cls: type[T_LocData],  # noqa: UP006
        dataframe: DataFrame | None = None,
        meta: (
            metadata_pb2.Metadata
            | dict[str, Any]
            | str
            | bytes
            | os.PathLike[Any]
            | BinaryIO
            | None
        ) = None,
    ) -> T_LocData:
        """
        Create new LocData object from DataFrame with localization data.

        Parameters
        ----------
        dataframe
            Localization data.
        meta
            Metadata about the current dataset and its history.

        Returns
        -------
        LocData
            A new LocData instance with dataframe representing the
            concatenated data.
        """
        dataframe = _dataframe_to_pandas(dataframe=dataframe, allow_copy=True)
        dataframe = pd.DataFrame() if dataframe is None else dataframe
        meta_ = metadata_pb2.Metadata()

        meta_.source = metadata_pb2.DESIGN
        meta_.state = metadata_pb2.RAW
        meta_.history.add(name="LocData.from_dataframe")

        meta_ = merge_metadata(metadata=meta_, other_metadata=meta)

        return cls(dataframe=dataframe, meta=meta_)

    @classmethod
    def from_coordinates(
        cls: type[T_LocData],  # noqa: UP006
        coordinates: npt.ArrayLike | None = None,
        coordinate_labels: Sequence[str] | None = None,
        meta: (
            metadata_pb2.Metadata
            | dict[str, Any]
            | str
            | bytes
            | os.PathLike[Any]
            | BinaryIO
            | None
        ) = None,
    ) -> T_LocData:
        """
        Create new LocData object from a sequence of localization coordinates.

        Parameters
        ----------
        coordinates
            Sequence of tuples with localization coordinates
            with shape (n_loclizations, dimension)
        coordinate_labels
            The available coordinate properties.
        meta
            Metadata about the current dataset and its history.

        Returns
        -------
        LocData
            A new LocData instance with dataframe representing the
            oncatenated data.
        """
        if coordinates is None:
            coordinates = np.array([])
        else:
            coordinates = np.asarray(coordinates)

        if np.size(coordinates):
            dimension = len(coordinates[0])

            if coordinate_labels is None:
                coordinate_labels = ["position_x", "position_y", "position_z"][
                    0:dimension
                ]
            else:
                if all(cl in PROPERTY_KEYS for cl in coordinate_labels):
                    coordinate_labels = coordinate_labels
                else:
                    raise ValueError(
                        "The given coordinate_keys are not standard property keys."
                    )

            dataframe = pd.DataFrame.from_records(
                data=coordinates, columns=coordinate_labels
            )

        else:
            dataframe = pd.DataFrame()

        meta_ = metadata_pb2.Metadata()
        meta_.source = metadata_pb2.DESIGN
        meta_.state = metadata_pb2.RAW
        meta_.history.add(name="LocData.from_coordinates")

        meta_ = merge_metadata(metadata=meta_, other_metadata=meta)

        return cls(dataframe=dataframe, meta=meta_)

    @classmethod
    def from_selection(
        cls: type[T_LocData],  # noqa: UP006
        locdata: LocData,
        indices: (
            int
            | list[int | bool]
            | npt.NDArray[np.int64 | np.bool_]
            | slice
            | pd.Index[int]
            | None
        ) = None,
        meta: (
            metadata_pb2.Metadata
            | dict[str, Any]
            | str
            | bytes
            | os.PathLike[Any]
            | BinaryIO
            | None
        ) = None,
    ) -> T_LocData:
        """
        Create new LocData object from selected elements in another `LocData`.

        Parameters
        ----------
        locdata
            Locdata object from which to select elements.
        indices
            Index labels for elements in locdata that make up the new data.
            Note that contrary to usual python slices, both the start and the
            stop are included (see pandas documentation).
            `Indices` refer to index value not position in list.
        meta
            Metadata about the current dataset and its history.

        Returns
        -------
        LocData
            A new LocData instance with dataframe representing the selected
            data.

        Note
        ----
        No error is raised if indices do not exist in locdata.
        """
        references = locdata
        if indices is None:
            indices = slice(0, None)

        meta_ = metadata_pb2.Metadata()
        meta_.CopyFrom(locdata.meta)
        try:
            meta_.ClearField("identifier")
        except ValueError:
            pass

        try:
            meta_.ClearField("element_count")
        except ValueError:
            pass

        try:
            meta_.ClearField("frame_count")
        except ValueError:
            pass

        meta_.modification_time.GetCurrentTime()
        meta_.state = metadata_pb2.MODIFIED
        meta_.ancestor_identifiers.append(locdata.meta.identifier)
        meta_.history.add(name="LocData.from_selection")

        meta_ = merge_metadata(metadata=meta_, other_metadata=meta)

        new_locdata = cls(references=references, indices=indices, meta=meta_)
        new_locdata.region = references.region
        return new_locdata

    @classmethod
    def from_collection(
        cls: type[T_LocData],  # noqa: UP006v
        locdatas: Iterable[LocData],
        meta: (
            metadata_pb2.Metadata
            | dict[str, Any]
            | str
            | bytes
            | os.PathLike[Any]
            | BinaryIO
            | None
        ) = None,
    ) -> T_LocData:
        """
        Create new LocData object by collecting LocData objects.

        Parameters
        ----------
        locdatas
            Locdata objects to collect.
        meta
            Metadata about the current dataset and its history.

        Returns
        -------
        LocData
            A new LocData instance with dataframe representing the concatenated data.
        """
        references = locdatas
        dataframe = pd.DataFrame([ref.properties for ref in references])

        meta_ = metadata_pb2.Metadata()

        meta_.source = metadata_pb2.DESIGN
        meta_.state = metadata_pb2.RAW
        meta_.ancestor_identifiers[:] = [ref.meta.identifier for ref in references]
        meta_.history.add(name="LocData.from_collection")

        meta_ = merge_metadata(metadata=meta_, other_metadata=meta)

        return cls(references=references, dataframe=dataframe, meta=meta_)

    @classmethod
    def concat(
        cls: type[T_LocData],  # noqa: UP006
        locdatas: Iterable[LocData],
        meta: (
            metadata_pb2.Metadata
            | dict[str, Any]
            | str
            | bytes
            | os.PathLike[Any]
            | BinaryIO
            | None
        ) = None,
    ) -> T_LocData:
        """
        Concatenate LocData objects.

        Parameters
        ----------
        locdatas
            Locdata objects to concatenate.
        meta
            Metadata about the current dataset and its history.

        Returns
        -------
        LocData
            A new LocData instance with dataframe representing the concatenated data.
        """

        dataframe = pd.concat([i.data for i in locdatas], ignore_index=True, sort=False)

        # concatenate references also if None
        references: list[LocData] = []
        for locdata in locdatas:
            try:
                references.extend(locdata.references)  # type: ignore
            except TypeError:
                references.append(locdata.references)  # type: ignore

        # check if all elements are None
        new_references = None if not any(references) else references

        meta_ = metadata_pb2.Metadata()

        meta_.creation_time.GetCurrentTime()
        meta_.source = metadata_pb2.DESIGN
        meta_.state = metadata_pb2.MODIFIED
        meta_.ancestor_identifiers[:] = [dat.meta.identifier for dat in locdatas]
        meta_.history.add(name="concat")

        meta_ = merge_metadata(metadata=meta_, other_metadata=meta)

        return cls(references=new_references, dataframe=dataframe, meta=meta_)

    @classmethod
    def from_chunks(
        cls: type[T_LocData],  # noqa: UP006
        locdata: LocData,
        chunks: Sequence[tuple[int, ...]] | None = None,
        chunk_size: int | None = None,
        n_chunks: int | None = None,
        order: Literal["successive", "alternating"] = "successive",
        drop: bool = False,
        meta: (
            metadata_pb2.Metadata
            | dict[str, Any]
            | str
            | bytes
            | os.PathLike[Any]
            | BinaryIO
            | None
        ) = None,
    ) -> T_LocData:
        """
        Divide locdata in chunks of localization elements.

        Parameters
        ----------
        locdata
            Locdata to divide.
        chunks
            Localization chunks as defined by a list of index-tuples.
            One of `chunks`, `chunk_size` or `n_chunks` must be different
            from None.
        chunk_size
            Number of consecutive localizations to form a single chunk of data.
            One of `chunks`, `chunk_size` or `n_chunks` must be different
            from None.
        n_chunks
            Number of chunks.
            One of `chunks`, `chunk_size` or `n_chunks` must be different
            from None.
        order
            The order in which to select localizations.
            One of 'successive' or 'alternating'.
        drop
            If True the last chunk will be eliminated if it has fewer
            localizations than the other chunks.
        meta
            Metadata about the current dataset and its history.

        Returns
        -------
        LocData
            A new LocData instance with references and dataframe elements
            representing the individual chunks.
        """
        n_nones = sum(element is None for element in [chunks, chunk_size, n_chunks])

        if n_nones != 2:
            raise ValueError(
                "One and only one of `chunks`, `chunk_size` or `n_chunks` must "
                "be different from None."
            )
        elif chunks is not None:
            index_lists = list(chunks)
        else:
            if chunk_size is not None:
                if (len(locdata) % chunk_size) == 0:
                    n_chunks = len(locdata) // chunk_size
                else:
                    n_chunks = len(locdata) // chunk_size + 1
            else:  # if n_chunks is not None
                assert n_chunks is not None  # type narrowing # noqa: S101
                if (len(locdata) % n_chunks) == 0:
                    chunk_size = len(locdata) // n_chunks
                else:
                    chunk_size = len(locdata) // (n_chunks - 1)

            if order == "successive":
                if (len(locdata) % chunk_size) == 0:
                    chunk_sizes = [chunk_size] * n_chunks
                else:
                    chunk_sizes = [chunk_size] * (n_chunks - 1) + [
                        (len(locdata) % chunk_size)
                    ]
                cum_chunk_sizes = list(accumulate(chunk_sizes))
                cum_chunk_sizes.insert(0, 0)
                index_lists = [
                    locdata.data.index[slice(lower, upper)]  # type: ignore
                    for lower, upper in zip(cum_chunk_sizes[:-1], cum_chunk_sizes[1:])
                ]

            elif order == "alternating":
                index_lists = [
                    locdata.data.index[slice(i_chunk, None, n_chunks)]  # type: ignore
                    for i_chunk in range(n_chunks)
                ]

            else:
                raise ValueError(f"The order {order} is not implemented.")

        if drop and len(index_lists) > 1 and len(index_lists[-1]) < len(index_lists[0]):
            index_lists = index_lists[:-1]

        references = [
            LocData.from_selection(locdata=locdata, indices=list(index_list))
            for index_list in index_lists
        ]
        dataframe = pd.DataFrame([ref.properties for ref in references])

        meta_ = metadata_pb2.Metadata()

        meta_.creation_time.GetCurrentTime()
        meta_.source = metadata_pb2.DESIGN
        meta_.state = metadata_pb2.RAW
        meta_.ancestor_identifiers[:] = [ref.meta.identifier for ref in references]
        meta_.history.add(name="LocData.chunks")

        meta_ = merge_metadata(metadata=meta_, other_metadata=meta)

        return cls(references=references, dataframe=dataframe, meta=meta_)

    def reset(self, reset_index: bool = False) -> Self:
        """
        Reset hulls and properties. This is needed after the dataframe
        attribute has been modified in place.

        Note
        ----
        Should be used with care because metadata is not updated accordingly.
        The region property is not changed.
        Better to just re-instantiate with `LocData.from_dataframe()` or
        use `locdata.update()`.

        Parameters
        ----------
        reset_index
            Flag indicating if the index is reset to integer values.
            If True the previous index values are discarded.

        Returns
        -------
        Self
            The modified object
        """
        if reset_index is True:
            self.dataframe.reset_index(drop=True, inplace=True)

        self.properties = {}
        self._bounding_box = None
        self._oriented_bounding_box = None
        self._convex_hull = None
        self._alpha_shape = None
        self._inertia_moments = None

        self._update_properties()

        return self

    def update(
        self,
        dataframe: pd.DataFrame | None,
        reset_index: bool = False,
        meta: (
            metadata_pb2.Metadata
            | dict[str, Any]
            | str
            | bytes
            | os.PathLike[Any]
            | BinaryIO
            | None
        ) = None,
    ) -> Self:
        """
        Update the dataframe attribute in place.

        Use this function rather than setting locdata.dataframe directly in
        order to automatically update
        the attributes for dimension, hulls, properties, and metadata.

        Parameters
        ----------
        dataframe
            Dataframe with localization data.
        reset_index
            Flag indicating if the index is reset to integer values.
            If True the previous index values are discarded.
        meta : locan.data.metadata_pb2.Metadata | dict | str | bytes | os.PathLike | BinaryIO | None
            Metadata about the current dataset and its history.

        Returns
        -------
        Self
            The modified object
        """
        if dataframe is None:
            return self

        local_parameter = locals()
        del local_parameter[
            "dataframe"
        ]  # dataframe is obvious and possibly large and should not be repeated in meta.

        if self.references is not None:
            self.reduce(reset_index=reset_index)
            logger.warning(
                "LocData.reduce() was applied since self.references was not None."
            )

        self.dataframe = dataframe
        self.dimension = len(self.coordinate_keys)
        self.reset(reset_index=reset_index)  # update hulls and properties

        # update meta
        self.meta.modification_time.GetCurrentTime()
        self.meta.state = metadata_pb2.MODIFIED
        self.meta.history.add(name="LocData.update", parameter=str(local_parameter))

        self.meta.element_count = len(self.data.index)
        if "frame" in self.data.columns:
            self.meta.frame_count = len(self.data["frame"].unique())

        self.meta = merge_metadata(metadata=self.meta, other_metadata=meta)

        return self

    def reduce(self, reset_index: bool = False) -> Self:
        """
        Clean up references.

        This includes to update `Locdata.dataframe` and set
        `LocData.references` and `LocData.indices` to None.

        Parameters
        ----------
        reset_index
            Flag indicating if the index is reset to integer values.
            If True the previous index values are discarded.

        Returns
        -------
        Self
            The modified object
        """
        if self.references is None:
            pass
        elif isinstance(self.references, (LocData, list)):
            self.dataframe = self.data
            self.indices = None
            self.references = None
        else:
            raise ValueError("references has undefined value.")

        if reset_index is True:
            self.dataframe.reset_index(drop=True, inplace=True)

        return self

    def update_convex_hulls_in_references(self) -> Self:
        """
        Compute the convex hull for each element in locdata.references and
        update locdata.dataframe.

        Returns
        -------
        Self
            The modified object
        """
        if isinstance(self.references, list):
            for reference in self.references:
                reference.convex_hull  # request property to update reference._convex_hull  # noqa B018

            new_df = pd.DataFrame(
                [reference.properties for reference in self.references]
            )
            new_df.index = self.data.index
            self.dataframe.update(new_df)
            new_columns = [
                column for column in new_df.columns if column in self.dataframe.columns
            ]
            new_df.drop(columns=new_columns, inplace=True, errors="ignore")
            self.dataframe = pd.concat([self.dataframe, new_df], axis=1)
        return self

    def update_oriented_bounding_box_in_references(self) -> Self:
        """
        Compute the oriented bounding box for each element in
        locdata.references and update locdata.dataframe.

        Returns
        -------
        Self
            The modified object
        """
        if isinstance(self.references, list):
            for reference in self.references:
                reference.oriented_bounding_box  # request property to update reference._convex_hull  # noqa B018
            new_df = pd.DataFrame(
                [reference.properties for reference in self.references]
            )
            new_df.index = self.data.index
            self.dataframe.update(new_df)
            new_columns = [
                column for column in new_df.columns if column in self.dataframe.columns
            ]
            new_df.drop(columns=new_columns, inplace=True, errors="ignore")
            self.dataframe = pd.concat([self.dataframe, new_df], axis=1)
        return self

    def projection(self, coordinate_labels: str | list[str]) -> LocData:
        """
        Reduce dimensions by projecting all localization coordinates onto
        selected coordinates.

        Parameters
        ----------
        coordinate_labels
            The coordinate labels to project onto.

        Returns
        -------
            LocData
        """
        local_parameter = locals()

        if isinstance(coordinate_labels, str):
            coordinate_labels = [coordinate_labels]

        new_locdata: LocData = copy.deepcopy(self)

        # reduce coordinate dimensions
        coordinate_labels_to_drop = [
            label for label in self.coordinate_keys if label not in coordinate_labels
        ]
        columns = self.data.columns
        new_columns = [
            column for column in columns if column not in coordinate_labels_to_drop
        ]
        dataframe = new_locdata.data[new_columns]

        # update
        _meta = metadata_pb2.Metadata()
        _meta.history.add(name="LocData.projection", parameter=str(local_parameter))
        # other updates are done in the coming update call.

        new_locdata = new_locdata.update(dataframe=dataframe, meta=_meta)

        return new_locdata

    def print_meta(self) -> None:
        """
        Print Locdata.metadata.

        See Also
        --------
        :func:`locan.data.metadata_utils.metadata_to_formatted_string`
        """
        print(metadata_to_formatted_string(self.meta))

    def print_summary(self) -> None:
        """
        Print a summary containing the most common metadata keys.
        """
        meta_ = metadata_pb2.Metadata()
        if self.meta.HasField("file"):
            meta_.file.CopyFrom(self.meta.file)
        meta_.identifier = self.meta.identifier
        meta_.comment = self.meta.comment
        meta_.creation_time.CopyFrom(self.meta.creation_time)
        if self.meta.HasField("modification_time"):
            meta_.modification_time.CopyFrom(self.meta.modification_time)
        meta_.source = self.meta.source
        meta_.state = self.meta.state
        meta_.element_count = self.meta.element_count
        meta_.frame_count = self.meta.frame_count

        print(metadata_to_formatted_string(meta_))

    def update_properties_in_references(
        self,
        properties: (
            dict[str, Iterable[Any]]
            | pd.Series[Any]
            | pd.DataFrame
            | Callable[..., Any]
            | None
        ) = None,
    ) -> Self:
        """
        Add properties for each element in self.references
        and update self.dataframe.

        Parameters
        ----------
        properties
            new property values for each reference or
            function to compute property for LocData object.

        Returns
        -------
        Self
        """
        if not isinstance(self.references, list):
            raise TypeError("self.references must be a list of Locdata")

        if properties is None:
            pass
        elif isinstance(properties, dict):
            for key, values in properties.items():
                for reference, value_ in zip(self.references, values):
                    reference.properties.update({key: value_})
        elif isinstance(properties, pd.Series):
            if any(properties.index != range(len(self.references))):
                raise ValueError(
                    f"properties must have a range index of length {len(self.references)}"
                )
            for reference, value_ in zip(self.references, properties.to_numpy()):
                reference.properties.update({properties.name: value_})
        elif isinstance(properties, pd.DataFrame):
            if any(properties.index != range(len(self.references))):
                raise ValueError(
                    f"properties must have a range index of length {len(self.references)}"
                )
            for name in properties.columns:
                for reference, value_ in zip(
                    self.references, properties[name].to_numpy()
                ):
                    reference.properties.update({name: value_})
        elif callable(properties):
            for reference in self.references:
                reference.properties.update(properties(reference))

        new_df = pd.DataFrame([reference.properties for reference in self.references])
        new_df.index = self.data.index

        references_ = self.references
        self.update(dataframe=new_df)
        self.references = references_

        return self



"""
Utility script to print locan.__all__ elements.

These items need to be explicitly re-exported in locan.__init__
to satisfy mypy.
"""

from importlib import import_module

submodules: list[str] = [
    "analysis",
    "configuration",
    "constants",
    "data",
    "datasets",
    "dependencies",
    "gui",
    "locan_io",
    "process",
    "rois",
    "scripts",
    "simulation",
    "tests",
    "utils",
    "visualize",
]


def make_init(submodule: str, package: str) -> str:
    """
    Replace
    'from .submodule import *'
    with
    'from .submodule import (<explicit list of functions and classes>)'
    according to provided __all__.

    The output will be copied into submodule.__init__.py
    or submodule.__init__.pyi.

    Parameters
    ----------
    submodule
        the module from which to import
    package
        the anchor point from which to resolve relative submodule specifications.

    Returns
    -------
    str
    """
    module = import_module(name=f".{submodule}", package=package)
    lines = ",\n".join([f"{item} as {item}" for item in sorted(module.__all__)])
    relative_submodule = ".".join(module.__name__.split("locan."))
    text = f"from {relative_submodule} import (\n" + lines + "\n)" + "\n"
    return text


def make_explicit_import_command(submodules: list[str], package: str) -> str:
    """
    Print explicitly re-exported imports from submodules
    'from .submodule import (<explicit list of functions and classes>)'
    according to provided __all__.

    The output will be copied into submodule.__init__.py
    or submodule.__init__.pyi.

    Parameters
    ----------
    submodules
        the module from which to import
    package
        the anchor point from which to resolve relative submodule specifications.

    Returns
    -------
    str
    """
    text = ""
    for submodule in submodules:
        module = import_module(name=f".{submodule}", package=package)
        lines = ",\n".join([f"{item} as {item}" for item in sorted(module.__all__)])
        relative_submodule = ".".join(module.__name__.split("locan."))
        text += f"from {relative_submodule} import (\n" + lines + "\n)" + "\n"
    return text


def make_explicit_all(submodules: list[str], package: str) -> str:
    """
    Print explicitly __all__
    according to provided submodules.__all__.

    The output will be copied into submodule.__init__.py
    or submodule.__init__.pyi.

    Parameters
    ----------
    submodules
        the module from which to import
    package
        the anchor point from which to resolve relative submodule specifications.

    Returns
    -------
    str
    """
    all_ = []
    for submodule in submodules:
        module = import_module(name=f".{submodule}", package="locan")
        all_.extend(module.__all__)
    text = "__all__ = [\n'" + "',\n'".join(sorted(all_)) + "',\n]" + "\n"
    return text


if __name__ == "__main__":
    package = "locan"
    print(submodules)
    # import locan as lc
    # print(lc.EXTRAS_REQUIRE)

    print(make_init(submodule="data.regions", package=package))

    # print(make_explicit_all(submodules=submodules, package=package))
    # print(make_explicit_import_command(submodules=submodules, package=package))



"""

Region of interest.

This module provides functions for managing regions of interest in
localization data.

Submodules:
-----------

.. autosummary::
   :toctree: ./

   roi
"""

from __future__ import annotations

from locan.rois import roi

from .roi import *

__all__: list[str] = []
__all__.extend(roi.__all__)



"""

File input/output functions.

This module provides functions for file input and output of data related to single-molecule localization microscopy.

Submodules:
-----------

.. autosummary::
   :toctree: ./

   files
   locdata
   utilities

"""

from __future__ import annotations

from locan.locan_io import files, locdata, utilities

from .files import *
from .locdata import (
    convert_property_names as convert_property_names,
    convert_property_types as convert_property_types,
    load_Elyra_file as load_Elyra_file,
    load_Elyra_header as load_Elyra_header,
    load_Nanoimager_file as load_Nanoimager_file,
    load_Nanoimager_header as load_Nanoimager_header,
    load_SMAP_file as load_SMAP_file,
    load_SMAP_header as load_SMAP_header,
    load_SMLM_file as load_SMLM_file,
    load_SMLM_header as load_SMLM_header,
    load_SMLM_manifest as load_SMLM_manifest,
    load_asdf_file as load_asdf_file,
    load_decode_file as load_decode_file,
    load_decode_header as load_decode_header,
    load_locdata as load_locdata,
    load_rapidSTORM_file as load_rapidSTORM_file,
    load_rapidSTORM_header as load_rapidSTORM_header,
    load_rapidSTORM_track_file as load_rapidSTORM_track_file,
    load_rapidSTORM_track_header as load_rapidSTORM_track_header,
    load_thunderstorm_file as load_thunderstorm_file,
    load_thunderstorm_header as load_thunderstorm_header,
    load_txt_file as load_txt_file,
    manifest_file_info_from_locdata as manifest_file_info_from_locdata,
    manifest_format_from_locdata as manifest_format_from_locdata,
    manifest_from_locdata as manifest_from_locdata,
    save_SMAP_csv as save_SMAP_csv,
    save_SMLM as save_SMLM,
    save_asdf as save_asdf,
    save_thunderstorm_csv as save_thunderstorm_csv,
)
from .utilities import *

__all__: list[str] = []
__all__.extend(files.__all__)
__all__.extend(locdata.__all__)
__all__.extend(utilities.__all__)



import matplotlib.pyplot as plt
import pytest

from locan import RenderEngine, render_2d, render_3d
from locan.dependencies import HAS_DEPENDENCY


@pytest.mark.gui
@pytest.mark.parametrize(
    "test_input, expected", list((member, 0) for member in list(RenderEngine))
)
def test_render_2d(locdata_blobs_2d, test_input, expected):
    if HAS_DEPENDENCY["napari"] and test_input == RenderEngine.NAPARI:
        render_2d(locdata_blobs_2d, render_engine=test_input)
        # napari.run()
    else:
        render_2d(locdata_blobs_2d, render_engine=test_input)
    # plt.show()

    plt.close("all")


@pytest.mark.gui
@pytest.mark.parametrize(
    "test_input, expected", list((member, 0) for member in list(RenderEngine))
)
def test_render_3d(locdata_blobs_3d, test_input, expected):
    if HAS_DEPENDENCY["napari"] and test_input == RenderEngine.NAPARI:
        render_3d(locdata_blobs_3d, render_engine=test_input)
        # napari.run()
    else:
        with pytest.raises(NotImplementedError):
            render_3d(locdata_blobs_3d, render_engine=test_input)
    # plt.show()

    plt.close("all")



"""
Benchmark functions for :func:`locan.data.hull`
"""

import locan as lc
from locan.data.hulls import AlphaShape, BoundingBox, ConvexHull, OrientedBoundingBox


class BenchmarkLocDataHulls:
    """
    Benchmarks for LocData hull computation
    """

    def setup(self):
        path = lc.ROOT_DIR / "tests/test_data/five_blobs.txt"
        self.locdata = lc.load_txt_file(path)
        # print(self.locdata.data.info())

    def check_bounding_box(self):
        bb = BoundingBox(points=self.locdata.coordinates)
        print(bb.region_measure)

    def time_bounding_box(self):
        BoundingBox(points=self.locdata.coordinates)

    def check_oriented_bounding_box(self):
        bb = OrientedBoundingBox(points=self.locdata.coordinates)
        print(bb.region_measure)

    def time_oriented_bounding_box(self):
        OrientedBoundingBox(points=self.locdata.coordinates)

    def check_convex_hull(self):
        bb = ConvexHull(points=self.locdata.coordinates)
        print(bb.region_measure)

    def time_convex_hull_scipy(self):
        ConvexHull(points=self.locdata.coordinates, method="scipy")

    def time_convex_hull_shapely(self):
        ConvexHull(points=self.locdata.coordinates, method="shapely")

    def check_alpha_shape(self):
        bb = AlphaShape(points=self.locdata.coordinates, alpha=100)
        print(bb.region_measure)

    def time_alpha_shape(self):
        AlphaShape(points=self.locdata.coordinates, alpha=100)


def main():
    bm = BenchmarkLocDataHulls()
    bm.setup()

    bm.check_bounding_box()
    bm.check_oriented_bounding_box()
    bm.check_convex_hull()
    bm.check_alpha_shape()

    bm.time_bounding_box()
    bm.time_oriented_bounding_box()
    bm.time_convex_hull_scipy()
    bm.time_convex_hull_shapely()
    bm.time_alpha_shape()


def main_profile():
    """
    Run benchmarks through profiler.
    """
    import cProfile
    import pstats

    bm = BenchmarkLocDataHulls()
    bm.setup()

    profiler = cProfile.Profile()
    profiler.enable()

    for _ in range(100):
        bm.time_bounding_box()
        bm.time_oriented_bounding_box()
        bm.time_convex_hull_scipy()
        bm.time_convex_hull_shapely()
        bm.time_alpha_shape()

    profiler.disable()
    stats = pstats.Stats(profiler).sort_stats("cumtime")
    stats.print_stats("time_")
    stats.print_stats()


if __name__ == "__main__":
    # main()
    main_profile()
